"""Tests the `_utils.base_type` module."""

from typing import Annotated, Literal, Union

from pydantic import BaseModel

from mirascope.core.base._utils._base_type import is_base_type


def test_is_base_type() -> None:
    """Tests the `is_base_type` TypeGuard function."""
    assert is_base_type(str)
    assert is_base_type(int)
    assert is_base_type(float)
    assert is_base_type(bool)
    assert is_base_type(list)
    assert is_base_type(set)
    assert is_base_type(tuple)
    assert is_base_type(dict)
    assert is_base_type(Union[str, int])  # noqa: UP007
    assert is_base_type(Annotated[str, "test"])
    assert is_base_type(Literal[1])
    assert not is_base_type(BaseModel)



from __future__ import annotations

from collections.abc import Sequence
from typing import TYPE_CHECKING

from pydantic import BaseModel
from typing_extensions import TypeIs

from .._utils._get_image_type import get_image_type
from ..message_param import (
    AudioPart,
    BaseMessageParam,
    CacheControlPart,
    ImagePart,
    TextPart,
)

if TYPE_CHECKING:
    from PIL import Image

    has_pil_module: bool
else:
    try:
        from PIL import Image  # pyright: ignore [reportAssignmentType]

        has_pil_module = True
    except ImportError:  # pragma: no cover
        has_pil_module = False

        class Image:
            class Image:
                def tobytes(self) -> bytes: ...


def _convert_message_sequence_part_to_content_part(
    message_sequence_part: str | Image.Image | TextPart | ImagePart | AudioPart,
) -> TextPart | ImagePart | AudioPart | CacheControlPart:
    if isinstance(message_sequence_part, str):
        return TextPart(text=message_sequence_part, type="text")
    elif isinstance(
        message_sequence_part, TextPart | ImagePart | AudioPart | CacheControlPart
    ):
        return message_sequence_part
    elif has_pil_module and isinstance(message_sequence_part, Image.Image):
        image = message_sequence_part.tobytes()
        return ImagePart(
            type="image",
            media_type=f"image/{get_image_type(image)}",
            image=image,
            detail=None,
        )
    else:
        raise ValueError(f"Invalid message sequence type: {message_sequence_part}")


def convert_message_content_to_message_param_content(
    message_sequence: Sequence[str | Image.Image | TextPart | ImagePart | AudioPart],
) -> list[TextPart | ImagePart | AudioPart | CacheControlPart] | str:
    if isinstance(message_sequence, str):
        return message_sequence
    return [
        _convert_message_sequence_part_to_content_part(message)
        for message in message_sequence
    ]


def _is_base_message_params(
    value: object,
) -> TypeIs[list[BaseMessageParam]]:
    # Note: we also need to catch the original provider message parameters here
    return isinstance(value, list) and all(
        isinstance(v, BaseMessageParam | dict | BaseModel) for v in value
    )


def convert_messages_to_message_params(
    messages: str
    | Sequence[str | Image.Image | TextPart | ImagePart | AudioPart]
    | list[BaseMessageParam]
    | BaseMessageParam,
    role: str = "user",
) -> list[BaseMessageParam]:
    if isinstance(messages, BaseMessageParam):
        return [messages]
    elif _is_base_message_params(messages):
        return messages
    elif isinstance(messages, str | Sequence):
        return [
            BaseMessageParam(
                content=convert_message_content_to_message_param_content(messages),
                role=role,
            )
        ]
    else:
        raise ValueError(f"Invalid messages type: {messages}")



import json

from mirascope.core import openai


@openai.call("gpt-4o-mini", json_mode=True)
def get_book_info(book_title: str) -> str:
    return f"Provide the author and genre of {book_title}"


try:
    response = get_book_info("The Name of the Wind")
    print(json.loads(response.content))
except json.JSONDecodeError:
    print("The model produced invalid JSON")



from __future__ import annotations

from collections.abc import Sequence
from typing import Any

from ._utils._convert_messages_to_message_params import (
    Image,
    convert_message_content_to_message_param_content,
)
from .message_param import AudioPart, BaseMessageParam, ImagePart, TextPart


class Messages:
    Type = (
        str
        | Sequence[str | Image.Image | TextPart | ImagePart | AudioPart]
        | list[BaseMessageParam | Any]
        | BaseMessageParam
    )

    @classmethod
    def System(
        cls,
        content: str | Sequence[str | Image.Image | TextPart | ImagePart | AudioPart],
    ) -> BaseMessageParam:
        return BaseMessageParam(
            role="system",
            content=convert_message_content_to_message_param_content(content),
        )

    @classmethod
    def User(
        cls,
        content: str | Sequence[str | Image.Image | TextPart | ImagePart | AudioPart],
    ) -> BaseMessageParam:
        return BaseMessageParam(
            role="user",
            content=convert_message_content_to_message_param_content(content),
        )

    @classmethod
    def Assistant(
        cls,
        content: str | Sequence[str | Image.Image | TextPart | ImagePart | AudioPart],
    ) -> BaseMessageParam:
        return BaseMessageParam(
            role="assistant",
            content=convert_message_content_to_message_param_content(content),
        )



from typing import Any, cast
from unittest.mock import Mock

import pytest

from mirascope.core import BaseMessageParam
from mirascope.core.base import BaseDynamicConfig
from mirascope.core.base._utils._get_dynamic_configuration import (
    get_dynamic_configuration,
)


def test_get_dynamic_configuration_sync_returns_config():
    def fn(*args: Any, **kwargs: Any) -> BaseDynamicConfig:
        return cast(BaseDynamicConfig, {"metadata": {"some_key": "some_value"}})

    result = get_dynamic_configuration(fn, (), {})
    assert result == {"metadata": {"some_key": "some_value"}}


def test_get_dynamic_configuration_sync_returns_message_param_list():
    def fn(*args: Any, **kwargs: Any) -> list[BaseMessageParam]:
        return [BaseMessageParam(role="user", content="Hello, World!")]

    result = get_dynamic_configuration(fn, (), {})
    assert result == {
        "messages": [BaseMessageParam(role="user", content="Hello, World!")]
    }


def test_get_dynamic_configuration_sync_with_original_fn():
    def original_fn(*args: Any, **kwargs: Any) -> BaseDynamicConfig:
        return cast(BaseDynamicConfig, {"computed_fields": {"field1": "value1"}})

    fn = Mock()
    fn._original_fn = original_fn

    result = get_dynamic_configuration(fn, (), {})
    assert result == {"computed_fields": {"field1": "value1"}}


def test_get_dynamic_configuration_sync_with_original_fn_returns_message_param_list():
    def original_fn(*args: Any, **kwargs: Any) -> list[BaseMessageParam]:
        return [BaseMessageParam(role="assistant", content="How can I assist you?")]

    fn = Mock()
    fn._original_fn = original_fn

    result = get_dynamic_configuration(fn, (), {})
    assert result == {
        "messages": [
            BaseMessageParam(role="assistant", content="How can I assist you?")
        ]
    }


def test_get_dynamic_configuration_sync_passes_args_kwargs():
    def fn(*args: Any, **kwargs: Any) -> BaseDynamicConfig:
        assert args == (1, 2)
        assert kwargs == {"key": "value"}
        return cast(
            BaseDynamicConfig,
            {"metadata": {"args_received": args, "kwargs_received": kwargs}},
        )

    result = get_dynamic_configuration(fn, (1, 2), {"key": "value"})
    assert result == {
        "metadata": {
            "args_received": (1, 2),
            "kwargs_received": {"key": "value"},
        }
    }


@pytest.mark.asyncio
async def test_get_dynamic_configuration_async_returns_config():
    async def fn(*args: Any, **kwargs: Any) -> BaseDynamicConfig:
        return cast(BaseDynamicConfig, {"metadata": {"some_key": "some_value"}})

    result = await get_dynamic_configuration(fn, (), {})
    assert result == {"metadata": {"some_key": "some_value"}}


@pytest.mark.asyncio
async def test_get_dynamic_configuration_async_returns_message_param_list():
    async def fn(*args: Any, **kwargs: Any) -> list[BaseMessageParam]:
        return [BaseMessageParam(role="user", content="Hello, World!")]

    result = await get_dynamic_configuration(fn, (), {})
    assert result == {
        "messages": [BaseMessageParam(role="user", content="Hello, World!")]
    }


@pytest.mark.asyncio
async def test_get_dynamic_configuration_async_with_original_fn():
    async def original_fn(*args: Any, **kwargs: Any) -> BaseDynamicConfig:
        return cast(BaseDynamicConfig, {"computed_fields": {"field1": "value1"}})

    fn = Mock()
    fn._original_fn = original_fn

    result = await get_dynamic_configuration(fn, (), {})
    assert result == {"computed_fields": {"field1": "value1"}}


@pytest.mark.asyncio
async def test_get_dynamic_configuration_async_with_original_fn_returns_message_param_list():
    async def original_fn(*args: Any, **kwargs: Any) -> list[BaseMessageParam]:
        return [BaseMessageParam(role="assistant", content="How can I assist you?")]

    fn = Mock()
    fn._original_fn = original_fn

    result = await get_dynamic_configuration(fn, (), {})
    assert result == {
        "messages": [
            BaseMessageParam(role="assistant", content="How can I assist you?")
        ]
    }


@pytest.mark.asyncio
async def test_get_dynamic_configuration_async_passes_args_kwargs():
    async def fn(*args: Any, **kwargs: Any) -> BaseDynamicConfig:
        assert args == (1, 2)
        assert kwargs == {"key": "value"}
        return cast(
            BaseDynamicConfig,
            {"metadata": {"args_received": args, "kwargs_received": kwargs}},
        )

    result = await get_dynamic_configuration(fn, (1, 2), {"key": "value"})
    assert result == {
        "metadata": {
            "args_received": (1, 2),
            "kwargs_received": {"key": "value"},
        }
    }



from mirascope.core import Messages, bedrock
from tenacity import retry, stop_after_attempt, wait_exponential


@retry(
    stop=stop_after_attempt(3),
    wait=wait_exponential(multiplier=1, min=4, max=10),
)
@bedrock.call("anthropic.claude-3-haiku-20240307-v1:0")
def recommend_book(genre: str) -> Messages.Type:
    return Messages.User(f"Recommend a {genre} book")


print(recommend_book("fantasy"))



from mirascope.core import bedrock


@bedrock.call("claude-3-5-sonnet-20240620")
def recommend_book(genre: str) -> bedrock.BedrockDynamicConfig:
    return {
        "messages": [
            {"role": "user", "content": [{"text": f"Recommend a {genre} book"}]}
        ]
    }


response = recommend_book("fantasy")
print(response.content)



from mirascope.core import mistral
from mistralai.models.chat_completion import ChatMessage


@mistral.call("mistral-large-latest")
def recommend_book(genre: str) -> mistral.MistralDynamicConfig:
    return {"messages": [ChatMessage(role="user", content=f"Recommend a {genre} book")]}


response = recommend_book("fantasy")
print(response.content)



"""Tests the `bedrock.tool` module."""

from mypy_boto3_bedrock_runtime.type_defs import (
    ToolSpecificationTypeDef,
    ToolTypeDef,
    ToolUseBlockOutputTypeDef,
)

from mirascope.core.base.tool import BaseTool
from mirascope.core.bedrock._types import ToolUseBlockContentTypeDef
from mirascope.core.bedrock.tool import BedrockTool, BedrockToolConfig


def test_bedrock_tool() -> None:
    """Tests the `BedrockTool` class."""

    class FormatBook(BedrockTool):
        """Returns the title and author nicely formatted."""

        title: str
        author: str

        tool_config = BedrockToolConfig()

        def call(self) -> str:
            return f"{self.title} by {self.author}"

    tool_call = ToolUseBlockContentTypeDef(
        toolUse=ToolUseBlockOutputTypeDef(
            toolUseId="id",
            input={"title": "The Name of the Wind", "author": "Patrick Rothfuss"},
            name="FormatBook",
        )
    )

    tool = FormatBook.from_tool_call(tool_call)
    assert isinstance(tool, BaseTool)
    assert isinstance(tool, BedrockTool)
    assert isinstance(tool, FormatBook)
    assert tool.title == "The Name of the Wind"
    assert tool.author == "Patrick Rothfuss"
    assert tool.call() == "The Name of the Wind by Patrick Rothfuss"

    expected_schema = ToolTypeDef(
        toolSpec=ToolSpecificationTypeDef(
            name="FormatBook",
            description="Returns the title and author nicely formatted.",
            inputSchema={
                "json": {
                    "properties": {
                        "title": {"type": "string"},
                        "author": {"type": "string"},
                    },
                    "required": ["title", "author"],
                    "type": "object",
                }
            },
        )
    )
    assert FormatBook.tool_schema() == expected_schema


def test_bedrock_tool_from_function() -> None:
    """Tests the `BedrockTool.type_from_fn` method."""

    def format_book(title: str, author: str) -> str:
        """Returns the title and author nicely formatted."""
        return f"{title} by {author}"

    ToolType = BedrockTool.type_from_fn(format_book)

    expected_schema = ToolTypeDef(
        toolSpec=ToolSpecificationTypeDef(
            name="format_book",
            description="Returns the title and author nicely formatted.",
            inputSchema={
                "json": {
                    "properties": {
                        "title": {"type": "string"},
                        "author": {"type": "string"},
                    },
                    "required": ["title", "author"],
                    "type": "object",
                }
            },
        )
    )
    assert ToolType.tool_schema() == expected_schema

    tool_call = ToolUseBlockContentTypeDef(
        toolUse={
            "toolUseId": "id",
            "input": {"title": "The Name of the Wind", "author": "Patrick Rothfuss"},
            "name": "format_book",
        }
    )
    tool = ToolType.from_tool_call(tool_call)
    assert tool.call() == "The Name of the Wind by Patrick Rothfuss"



from mirascope.core import litellm, prompt_template


def parse_recommendation(response: litellm.OpenAICallResponse) -> tuple[str, str]:
    title, author = response.content.split(" by ")
    return (title, author)


@litellm.call("gpt-4o-mini", output_parser=parse_recommendation)
@prompt_template("Recommend a {genre} book. Output only Title by Author")
def recommend_book(genre: str): ...


print(recommend_book("fantasy"))
# Output: ('"The Name of the Wind"', 'Patrick Rothfuss')



from mirascope.core import vertex
from vertexai.generative_models import Content, Part


@vertex.call("gemini-1.5-flash")
def recommend_book(genre: str) -> vertex.VertexDynamicConfig:
    return {
        "messages": [
            Content(role="user", parts=[Part.from_text(f"Recommend a {genre} book")])
        ]
    }


response = recommend_book("fantasy")
print(response.content)



from mirascope.core import gemini
from tenacity import retry, stop_after_attempt, wait_exponential


@retry(
    stop=stop_after_attempt(3),
    wait=wait_exponential(multiplier=1, min=4, max=10),
)
@gemini.call("gemini-1.5-flash")
def recommend_book(genre: str) -> str:
    return f"Recommend a {genre} book"


print(recommend_book("fantasy"))



from anthropic import AnthropicError
from mirascope.core import anthropic, prompt_template


@anthropic.call("claude-3-5-sonnet-20240620")
@prompt_template("Recommend a {genre} book")
def recommend_book(genre: str): ...


try:
    response = recommend_book("fantasy")
    print(response.content)
except AnthropicError as e:
    print(f"Error: {str(e)}")



"""Utitlies for more easily using Tenacity to reinsert errors on failed API calls."""

from collections.abc import Callable

from tenacity import RetryCallState


def collect_errors(
    *args: type[Exception],
) -> Callable[[RetryCallState], None]:
    """Collects specified errors into an `errors` keyword argument.

    Example:

    ```python
    from mirascope.integrations.tenacity import collect_errors
    from tenacity import retry, stop_after_attempt

    @retry(stop=stop_after_attempt(3), after=collect_errors(ValueError))
    def throw_value_error(*, errors: list[ValueError] | None = None):
        if errors:
            print(errors[-1])
        raise ValueError("Throwing Error")


    try:
        throw_value_error()
        # > Throwing Error
        # > Throwing Error
    except RetryError:
        ...
    ```

    Args:
        *args: The `Exception` types to catch.
    """

    def inner(retry_state: RetryCallState) -> None:
        """Collect errors into a `errors` kwarg of the wrapped function."""
        if (
            (outcome := retry_state.outcome)
            and (exception := outcome.exception())
            and type(exception) in args
        ):
            errors = retry_state.kwargs.pop("errors", [])
            retry_state.kwargs["errors"] = (
                errors + [exception] if errors else [exception]
            )

    return inner



from openai import OpenAI
from pydantic import BaseModel

client = OpenAI()


class Book(BaseModel):
    """An extracted book."""

    title: str
    author: str


def extract_book(text: str) -> Book:
    completion = client.chat.completions.create(
        model="gpt-4o-mini",
        messages=[{"role": "user", "content": f"Extract {text}"}],
        tools=[
            {
                "function": {
                    "name": "Book",
                    "description": "An extracted book.",
                    "parameters": {
                        "properties": {
                            "title": {"type": "string"},
                            "author": {"type": "string"},
                        },
                        "required": ["title", "author"],
                        "type": "object",
                    },
                },
                "type": "function",
            }
        ],
        tool_choice="required",
    )
    if tool_calls := completion.choices[0].message.tool_calls:
        return Book.model_validate_json(tool_calls[0].function.arguments)
    raise ValueError("No tool call found")


book = extract_book("The Name of the Wind by Patrick Rothfuss")
print(book)
# Output: title='The Name of the Wind' author='Patrick Rothfuss'



"""Tests the `azure.call_response` module."""

from datetime import datetime

from azure.ai.inference.models import (
    ChatChoice,
    ChatCompletions,
    ChatCompletionsToolCall,
    ChatResponseMessage,
    CompletionsFinishReason,
    CompletionsUsage,
    FunctionCall,
    ToolMessage,
)

from mirascope.core.azure.call_response import AzureCallResponse
from mirascope.core.azure.tool import AzureTool


def test_azure_call_response() -> None:
    """Tests the `AzureCallResponse` class."""
    choices = [
        ChatChoice(
            finish_reason="stop",
            index=0,
            message=ChatResponseMessage(**{"content": "content", "role": "assistant"}),  # pyright: ignore [reportCallIssue, reportArgumentType]
        )
    ]
    usage = CompletionsUsage(completion_tokens=1, prompt_tokens=1, total_tokens=2)
    completion = ChatCompletions(
        id="id",
        choices=choices,
        created=datetime.fromtimestamp(0),
        model="gpt-4o",
        usage=usage,
    )
    call_response = AzureCallResponse(
        metadata={},
        response=completion,
        tool_types=None,
        prompt_template="",
        fn_args={},
        dynamic_config=None,
        messages=[],
        call_params={},
        call_kwargs={},
        user_message_param=None,
        start_time=0,
        end_time=0,
    )
    assert call_response._provider == "azure"
    assert call_response.content == "content"
    assert call_response.finish_reasons == ["CompletionsFinishReason.STOPPED"]
    assert call_response.model == "gpt-4o"
    assert call_response.id == "id"
    assert call_response.usage == usage
    assert call_response.input_tokens == 1
    assert call_response.output_tokens == 1
    assert call_response.cost is None
    assert call_response.message_param == {
        "content": "content",
        "role": "assistant",
    }
    assert call_response.tools is None
    assert call_response.tool is None


def test_azure_call_response_with_tools() -> None:
    """Tests the `AzureCallResponse` class with tools."""

    class FormatBook(AzureTool):
        title: str
        author: str

        def call(self) -> str:
            return f"{self.title} by {self.author}"

    tool_call = ChatCompletionsToolCall(
        id="id",
        function=FunctionCall(
            arguments='{"title": "The Name of the Wind", "author": "Patrick Rothfuss"}',
            name="FormatBook",
        ),
    )
    completion = ChatCompletions(
        id="id",
        choices=[
            ChatChoice(
                finish_reason=CompletionsFinishReason("stop"),
                index=0,
                message=ChatResponseMessage(
                    content="content", tool_calls=[tool_call], role="assistant"
                ),
            )
        ],
        created=datetime.fromtimestamp(0),
        model="gpt-4o",
        usage=CompletionsUsage(completion_tokens=1, prompt_tokens=1, total_tokens=2),
    )
    call_response = AzureCallResponse(
        metadata={},
        response=completion,
        tool_types=[FormatBook],
        prompt_template="",
        fn_args={},
        dynamic_config=None,
        messages=[],
        call_params={},
        call_kwargs={},
        user_message_param=None,
        start_time=0,
        end_time=0,
    )
    tools = call_response.tools
    tool = call_response.tool
    assert tools and len(tools) == 1 and tools[0] == tool
    assert isinstance(tool, FormatBook)
    assert tool.title == "The Name of the Wind"
    assert tool.author == "Patrick Rothfuss"
    output = tool.call()
    assert output == "The Name of the Wind by Patrick Rothfuss"
    tool_message = ToolMessage(
        content=output,
        tool_call_id=tool_call.id,
    )
    tool_message.name = "FormatBook"  # type: ignore
    assert call_response.tool_message_params([(tool, output)]) == [tool_message]



from mirascope.core import BaseTool, openai
from mirascope.core.openai import OpenAIToolConfig


class FormatBook(BaseTool):
    title: str
    author: str

    tool_config = OpenAIToolConfig(strict=True)

    def call(self) -> str:
        return f"{self.title} by {self.author}"


@openai.call(
    "gpt-4o-2024-08-06", tools=[FormatBook], call_params={"tool_choice": "required"}
)
def recommend_book(genre: str) -> str:
    return f"Recommend a {genre} book"


response = recommend_book("fantasy")
if tool := response.tool:
    print(tool.call())



from mirascope.core import BaseMessageParam, openai
from openai import OpenAIError


@openai.call("gpt-4o-mini")
def recommend_book(genre: str) -> list[BaseMessageParam]:
    return [BaseMessageParam(role="user", content=f"Recommend a {genre} book")]


try:
    response = recommend_book("fantasy")
    print(response.content)
except OpenAIError as e:
    print(f"Error: {str(e)}")



from cohere.errors import BadRequestError
from mirascope.core import cohere


@cohere.call(model="command-r-plus")
def recommend_book(genre: str) -> str:
    return f"Recommend a {genre} book"


try:
    response = recommend_book("fantasy")
    print(response.content)
except BadRequestError as e:
    print(f"Error: {str(e)}")



import asyncio

from mirascope.core import gemini, prompt_template


@gemini.call("gemini-1.5-flash")
@prompt_template("Recommend a {genre} book")
async def recommend_book(genre: str): ...


async def main():
    genres = ["fantasy", "scifi", "mystery"]
    tasks = [recommend_book(genre) for genre in genres]
    results = await asyncio.gather(*tasks)

    for genre, response in zip(genres, results):
        print(f"({genre}):\n{response.content}\n")


asyncio.run(main())



from google.generativeai import GenerativeModel
from mirascope.core import Messages, gemini


@gemini.call("", client=GenerativeModel(model_name="gemini-1.5-flash"))
def recommend_book(genre: str) -> Messages.Type:
    return Messages.User(f"Recommend a {genre} book")



from mirascope.core import bedrock
from mirascope.integrations.otel import configure, with_otel

configure()


@with_otel()
@bedrock.call("anthropic.claude-3-haiku-20240307-v1:0")
def recommend_book(genre: str) -> str:
    return f"Recommend a {genre} book."


print(recommend_book("fantasy"))



from mirascope.core import Messages, vertex
from vertexai.generative_models import GenerationConfig


@vertex.call(
    "gemini-1.5-flash",
    call_params={"generation_config": GenerationConfig(max_output_tokens=512)},
)
def recommend_book(genre: str) -> Messages.Type:
    return Messages.User(f"Recommend a {genre} book")


response = recommend_book("fantasy")
print(response.content)



"""This module contains the `VertexCallResponseChunk` class.

usage docs: learn/streams.md#handling-streamed-responses
"""

from vertexai.generative_models import FinishReason, GenerationResponse

from ..base import BaseCallResponseChunk


class VertexCallResponseChunk(
    BaseCallResponseChunk[GenerationResponse, FinishReason]  # , Candidate.FinishReason]
):
    """A convenience wrapper around the Vertex AI streamed response chunks.

    When calling the Vertex AI API using a function decorated with `vertex_call` and
    `stream` set to `True`, the stream will contain `VertexCallResponseChunk` instances
    with properties that allow for more convenient access to commonly used attributes.

    Example:

    ```python
    from mirascope.core import prompt_template
    from mirascope.core.vertex import vertex_call


    @vertex_call("gemini-1.5-flash", stream=True)
    def recommend_book(genre: str) -> str:
        return f"Recommend a {genre} book"

    stream = recommend_book("fantasy")  # response is an `VertexStream`
    for chunk, _ in stream:
        print(chunk.content, end="", flush=True)
    ```
    """

    @property
    def content(self) -> str:
        """Returns the chunk content for the 0th choice."""
        return self.chunk.candidates[0].content.parts[0].text

    @property
    def finish_reasons(self) -> list[FinishReason]:
        """Returns the finish reasons of the response."""
        return [candidate.finish_reason for candidate in self.chunk.candidates]

    @property
    def model(self) -> None:
        """Returns the model name.

        vertex does not return model, so we return None
        """
        return None

    @property
    def id(self) -> str | None:
        """Returns the id of the response.

        vertex does not return an id
        """
        return None

    @property
    def usage(self) -> None:
        """Returns the usage of the chat completion.

        vertex does not have Usage, so we return None
        """
        return None

    @property
    def input_tokens(self) -> None:
        """Returns the number of input tokens."""
        return None

    @property
    def output_tokens(self) -> None:
        """Returns the number of output tokens."""
        return None



import asyncio

from mirascope.core import BaseMessageParam, openai


@openai.call(model="gpt-4o-mini")
async def recommend_book(genre: str) -> list[BaseMessageParam]:
    return [BaseMessageParam(role="user", content=f"Recommend a {genre} book")]


async def main():
    response = await recommend_book("fantasy")
    print(response.content)


asyncio.run(main())



"""This module defines the function return type for functions as LLM calls."""

from azure.ai.inference.models import ChatRequestMessage

from ..base import BaseDynamicConfig, BaseMessageParam
from .call_params import AzureCallParams

AzureDynamicConfig = BaseDynamicConfig[
    ChatRequestMessage | BaseMessageParam, AzureCallParams
]
"""The function return type for functions wrapped with the `azure_call` decorator.

Example:

```python
from mirascope.core import prompt_template
from mirascope.core.azure import AzureDynamicConfig, azure_call


@azure_call("gpt-4o-mini")
@prompt_template("Recommend a {capitalized_genre} book")
def recommend_book(genre: str) -> AzureDynamicConfig:
    return {"computed_fields": {"capitalized_genre": genre.capitalize()}}
```
"""



from cohere import Client

client = Client()


def recommend_book(genre: str) -> str:
    response = client.chat(
        model="command-r-plus",
        message=f"Recommend a {genre} book",
    )
    return response.text


output = recommend_book("fantasy")
print(output)



from mirascope.core import azure
from mirascope.integrations.langfuse import with_langfuse


@with_langfuse()
@azure.call("gpt-4o-mini")
def recommend_book(genre: str) -> str:
    return f"Recommend a {genre} book"


print(recommend_book("fantasy"))



from mirascope.core import Messages, litellm


@litellm.call("gpt-4o-mini")
def summarize(text: str) -> Messages.Type:
    return Messages.User(f"Summarize this text: {text}")


@litellm.call("gpt-4o-mini")
def translate(text: str, language: str) -> Messages.Type:
    return Messages.User(f"Translate this text to {language}: {text}")


summary = summarize("Long English text here...")
translation = translate(summary.content, "french")
print(translation.content)



from azure.ai.inference import ChatCompletionsClient
from azure.ai.inference.models import (
    ChatCompletionsToolDefinition,
    ChatRequestMessage,
    FunctionDefinition,
)
from azure.core.credentials import AzureKeyCredential
from pydantic import BaseModel

client = ChatCompletionsClient(
    endpoint="YOUR_ENDPOINT", credential=AzureKeyCredential("YOUR_KEY")
)


class Book(BaseModel):
    """An extracted book."""

    title: str
    author: str


def extract_book(text: str) -> Book:
    completion = client.complete(
        model="gpt-4o-mini",
        messages=[ChatRequestMessage({"role": "user", "content": f"Extract {text}"})],
        tools=[
            ChatCompletionsToolDefinition(
                function=FunctionDefinition(
                    name="Book",
                    description="An extracted book.",
                    parameters={
                        "properties": {
                            "title": {"type": "string"},
                            "author": {"type": "string"},
                        },
                        "required": ["title", "author"],
                        "type": "object",
                    },
                )
            )
        ],
        tool_choices="required",
    )
    if tool_calls := completion.choices[0].message.tool_calls:
        return Book.model_validate_json(tool_calls[0].function.arguments)
    raise ValueError("No tool call found")


book = extract_book("The Name of the Wind by Patrick Rothfuss")
print(book)
# Output: title='The Name of the Wind' author='Patrick Rothfuss'



"""Tests the `anthropic._utils.handle_stream` module."""

from unittest.mock import MagicMock

import pytest
from anthropic.types import (
    Message,
    MessageDeltaUsage,
    MessageStreamEvent,
    RawContentBlockDeltaEvent,
    RawContentBlockStartEvent,
    RawContentBlockStopEvent,
    RawMessageDeltaEvent,
    RawMessageStartEvent,
    RawMessageStopEvent,
    TextBlock,
    TextDelta,
    ToolUseBlock,
    Usage,
)

try:
    from anthropic.types import (
        InputJsonDelta as InputJSONDelta,  # pyright: ignore [reportAttributeAccessIssue]
    )
except ImportError:
    from anthropic.types import (
        InputJSONDelta,  # pyright: ignore [reportAttributeAccessIssue]
    )

from anthropic.types.raw_message_delta_event import Delta

from mirascope.core.anthropic._utils._handle_stream import (
    _handle_chunk,
    handle_stream,
    handle_stream_async,
)
from mirascope.core.anthropic.tool import AnthropicTool


class FormatBook(AnthropicTool):
    """Returns the title and author nicely formatted."""

    title: str
    author: str

    def call(self) -> None:
        """Dummy call."""


@pytest.fixture()
def mock_chunks() -> list[MessageStreamEvent]:
    """Returns a list of mock `MessageStreamEvent` instances."""
    return [
        RawMessageStartEvent(
            message=Message(
                id="msg_01UqKdZkJHmGJJRpNLMsBeAL",
                content=[],
                model="claude-3-5-sonnet-20240620",
                role="assistant",
                stop_reason=None,
                stop_sequence=None,
                type="message",
                usage=Usage(input_tokens=574, output_tokens=3),
            ),
            type="message_start",
        ),
        RawContentBlockStartEvent(
            content_block=TextBlock(text="", type="text"),
            index=0,
            type="content_block_start",
        ),
        RawContentBlockDeltaEvent(
            delta=TextDelta(text="Certainly! I can help", type="text_delta"),
            index=0,
            type="content_block_delta",
        ),
        RawContentBlockStopEvent(index=0, type="content_block_stop"),
        RawContentBlockStartEvent(
            content_block=ToolUseBlock(
                id="id",
                input={},
                name="FormatBook",
                type="tool_use",
            ),
            index=1,
            type="content_block_start",
        ),
        RawContentBlockDeltaEvent(
            delta=InputJSONDelta(
                partial_json='{"title": "The Name of the Wind", "author": "Patrick Rothfuss"}',
                type="input_json_delta",
            ),
            index=1,
            type="content_block_delta",
        ),
        RawContentBlockStopEvent(index=1, type="content_block_stop"),
        RawMessageDeltaEvent(
            delta=Delta(stop_reason="tool_use", stop_sequence=None),
            type="message_delta",
            usage=MessageDeltaUsage(output_tokens=127),
        ),
        RawMessageStopEvent(type="message_stop"),
    ]


def test_handle_stream(mock_chunks: list[MessageStreamEvent]) -> None:
    """Tests the `handle_stream` function."""

    result = list(handle_stream((c for c in mock_chunks), tool_types=[FormatBook]))
    assert len(result) == 9
    assert result[0][1] is None
    assert (
        (tool := result[6][1]) is not None
        and isinstance(tool, FormatBook)
        and tool.model_dump(exclude={"tool_call"})
        == {"title": "The Name of the Wind", "author": "Patrick Rothfuss"}
    )


@pytest.mark.asyncio
async def test_handle_stream_async(mock_chunks: list[MessageStreamEvent]) -> None:
    """Tests the `handle_stream_async` function."""

    async def generator():
        for chunk in mock_chunks:
            yield chunk

    result = []
    async for t in handle_stream_async(generator(), tool_types=[FormatBook]):
        result.append(t)
    assert len(result) == 9
    assert result[0][1] is None
    assert (
        (tool := result[6][1]) is not None
        and isinstance(tool, FormatBook)
        and tool.model_dump(exclude={"tool_call"})
        == {"title": "The Name of the Wind", "author": "Patrick Rothfuss"}
    )


def test_handle_chunk_no_tool_types() -> None:
    """Tests the `_handle_chunk` function with no tool types."""
    mock_chunk = MagicMock(spec=MessageStreamEvent)
    mock_current_tool_call = MagicMock(spec=ToolUseBlock)
    buffer, chunk, current_tool_call, current_tool_type = _handle_chunk(
        "",
        mock_chunk,
        mock_current_tool_call,
        None,
        None,
    )
    assert buffer == ""
    assert chunk is None
    assert current_tool_call == mock_current_tool_call
    assert current_tool_type is None



import inspect

from mirascope.core import BaseMessageParam, prompt_template


@prompt_template()
def recommend_book_prompt(genre: str) -> list[BaseMessageParam]:
    return [
        BaseMessageParam(
            role="user",
            content=inspect.cleandoc(
                f"""
                Recommend a {genre} book.
                Output in the format Title by Author.
                """
            ),
        ),
    ]


print(recommend_book_prompt("fantasy"))
# Output: [BaseMessageParam(role='system', content='Recommend a fantasy book.\nOutput in the format Title by Author.')]



"""Tests that `partial` works to make all fields optional."""

from pydantic import BaseModel

from mirascope.core.base._partial import partial


class ShallowModel(BaseModel):
    """A test model."""

    param: str
    default: int = 0


class PartialShallowModel(BaseModel):
    """A test model."""

    param: str | None = None
    default: int | None = None


def test_shallow_partial() -> None:
    """Tests that `Partial` works to make all fields optional in a shallow model."""
    assert (
        partial(ShallowModel).model_json_schema()
        == PartialShallowModel.model_json_schema()
    )


class DeeperModel(BaseModel):
    """A deeper model."""

    shallow: ShallowModel
    param: str


class PartialDeeperModel(BaseModel):
    """A deeper model."""

    shallow: PartialShallowModel | None = {}  # pyright: ignore [reportAssignmentType]
    param: str | None = None


def test_deeper_partial() -> None:
    """Tests that `Partial` works to make all fields optional in a deeper model."""
    assert (
        partial(DeeperModel).model_json_schema()
        == PartialDeeperModel.model_json_schema()
    )


class DeepestModel(BaseModel):
    """A deepest model."""

    shallow: ShallowModel
    deeper: DeeperModel
    param: str


class PartialDeepestModel(BaseModel):
    """A deepest model."""

    shallow: PartialShallowModel | None = {}  # pyright: ignore [reportAssignmentType]
    deeper: PartialDeeperModel | None = {}  # pyright: ignore [reportAssignmentType]
    param: str | None = None


def test_deepest_partial() -> None:
    """Tests that `Partial` works to make all fields optional in a deeper model."""
    assert (
        partial(DeepestModel).model_json_schema()
        == PartialDeepestModel.model_json_schema()
    )



"""Get FromCallArgs() annotated fields from a function call's args/kwargs."""

import inspect
from collections.abc import Callable
from typing import Any

from pydantic import BaseModel

from mirascope.core.base._utils import get_fn_args
from mirascope.core.base.from_call_args import is_from_call_args


def get_fields_from_call_args(
    response_model: object,
    fn: Callable,
    args: tuple[object, ...],
    kwargs: dict[str, Any],
) -> dict[str, Any]:
    if not (inspect.isclass(response_model) and issubclass(response_model, BaseModel)):
        return {}
    call_args_fields = {
        name
        for name, field in response_model.model_fields.items()
        if is_from_call_args(field)
    }

    fn_args = get_fn_args(fn, args, kwargs)
    if not call_args_fields.issubset(fn_args.keys()):
        raise ValueError(
            f"The function arguments do not contain all the fields marked with `FromCallArgs`. {fn_args=}, {call_args_fields=}"
        )
    return {name: fn_args[name] for name in call_args_fields}



from mirascope.core import cohere, prompt_template


def get_book_author(title: str) -> str:
    """Returns the author of the book with the given title

    Example:
        {"title": "The Name of the Wind"}

    Args:
        title: The title of the book.
    """
    if title == "The Name of the Wind":
        return "Patrick Rothfuss"
    elif title == "Mistborn: The Final Empire":
        return "Brandon Sanderson"
    else:
        return "Unknown"


@cohere.call("command-r-plus", tools=[get_book_author])
@prompt_template("Who wrote {book}?")
def identify_author(book: str): ...


response = identify_author("The Name of the Wind")
if tool := response.tool:
    print(tool.call())
else:
    print(response.content)



"""usage docs: learn/calls.md#provider-specific-parameters"""

from __future__ import annotations

from mistralai.models.chat_completion import ResponseFormat, ToolChoice
from typing_extensions import NotRequired

from ..base import BaseCallParams


class MistralCallParams(BaseCallParams):
    """The parameters to use when calling the Mistral API.

    [Mistral API Reference](https://docs.mistral.ai/api/)

    Attributes:
        endpoint: ...
        max_tokens: ...
        random_seed: ...
        response_format: ...
        safe_mode: ...
        safe_prompt: ...
        temperature: ...
        tool_choice: ...
        top_p: ...
    """

    endpoint: NotRequired[str | None]
    max_tokens: NotRequired[int | None]
    random_seed: NotRequired[int | None]
    response_format: NotRequired[ResponseFormat | None]
    safe_mode: NotRequired[bool | None]
    safe_prompt: NotRequired[bool | None]
    temperature: NotRequired[float | None]
    tool_choice: NotRequired[ToolChoice | None]
    top_p: NotRequired[float | None]



"""This module contains the setup_call function for OpenAI tools."""

from collections.abc import Awaitable, Callable
from typing import Any, cast, overload

from litellm import acompletion, completion
from openai import OpenAI
from openai.types.chat import ChatCompletion, ChatCompletionMessageParam

from ...base import BaseTool
from ...base._utils import fn_is_async
from ...openai import OpenAICallParams, OpenAIDynamicConfig, OpenAITool
from ...openai._utils import setup_call as setup_call_openai
from ...openai.call_kwargs import OpenAICallKwargs


@overload
def setup_call(
    *,
    model: str,
    client: None,
    fn: Callable[..., Awaitable[OpenAIDynamicConfig]],
    fn_args: dict[str, Any],
    dynamic_config: OpenAIDynamicConfig,
    tools: list[type[BaseTool] | Callable] | None,
    json_mode: bool,
    call_params: OpenAICallParams,
    extract: bool,
) -> tuple[
    Callable[..., Awaitable[ChatCompletion]],
    str | None,
    list[ChatCompletionMessageParam],
    list[type[OpenAITool]] | None,
    OpenAICallKwargs,
]: ...


@overload
def setup_call(
    *,
    model: str,
    client: None,
    fn: Callable[..., OpenAIDynamicConfig],
    fn_args: dict[str, Any],
    dynamic_config: OpenAIDynamicConfig,
    tools: list[type[BaseTool] | Callable] | None,
    json_mode: bool,
    call_params: OpenAICallParams,
    extract: bool,
) -> tuple[
    Callable[..., ChatCompletion],
    str | None,
    list[ChatCompletionMessageParam],
    list[type[OpenAITool]] | None,
    OpenAICallKwargs,
]: ...


def setup_call(
    *,
    model: str,
    client: None,
    fn: Callable[..., OpenAIDynamicConfig | Awaitable[OpenAIDynamicConfig]],
    fn_args: dict[str, Any],
    dynamic_config: OpenAIDynamicConfig,
    tools: list[type[BaseTool] | Callable] | None,
    json_mode: bool,
    call_params: OpenAICallParams,
    extract: bool,
) -> tuple[
    Callable[..., ChatCompletion] | Callable[..., Awaitable[ChatCompletion]],
    str | None,
    list[ChatCompletionMessageParam],
    list[type[OpenAITool]] | None,
    OpenAICallKwargs,
]:
    _, prompt_template, messages, tool_types, call_kwargs = setup_call_openai(
        model=model,  # pyright: ignore [reportCallIssue]
        client=OpenAI(api_key="NOT_USED"),
        fn=fn,  # pyright: ignore [reportArgumentType]
        fn_args=fn_args,  # pyright: ignore [reportArgumentType]
        dynamic_config=dynamic_config,
        tools=tools,
        json_mode=json_mode,
        call_params=call_params,
        extract=extract,
    )
    create = cast(
        Callable[..., ChatCompletion] | Callable[..., Awaitable[ChatCompletion]],
        acompletion if fn_is_async(fn) else completion,
    )
    return create, prompt_template, messages, tool_types, call_kwargs



from vertexai.generative_models import GenerativeModel

client = GenerativeModel("gemini-1.5-flash")


def recommend_book(genre: str) -> str:
    generation = client.generate_content(
        contents=[{"role": "user", "parts": f"Recommend a {genre} book"}]
    )
    return generation.candidates[0].content.parts[0].text  # type: ignore


output = recommend_book("fantasy")
print(output)



from mirascope.core import mistral, prompt_template
from mirascope.integrations.langfuse import with_langfuse


@with_langfuse()
@mistral.call("mistral-large-latest")
@prompt_template("Recommend a {genre} book")
def recommend_book(genre: str): ...


print(recommend_book("fantasy"))



from mirascope.core import cohere


@cohere.call("command-r-plus")
def recommend_book(genre: str) -> str:
    return f"Recommend a {genre} book"


response = recommend_book("fantasy")
print(response.content)



from mirascope.core import BaseMessageParam
from mirascope.core.base import AudioPart, TextPart


def identify_book_prompt(audio_mp3: bytes) -> list[BaseMessageParam]:
    return [
        BaseMessageParam(
            role="user",
            content=[
                TextPart(type="text", text="Here's an audio book snippet:"),
                AudioPart(
                    type="audio",
                    media_type="audio/mp3",
                    audio=audio_mp3,
                ),
                TextPart(type="text", text="What book is this?"),
            ],
        )
    ]


print(identify_book_prompt(b"..."))
# Output: [
#     BaseMessageParam(
#         role="user",
#         content=[
#             TextPart(type="text", text="Here's an audio book snippet:"),
#             ImagePart(type="image", media_type="image/jpeg", image=b"...", detail=None),
#             TextPart(type="text", text="What book is this?"),
#         ],
#     )
# ]



from mirascope.core import Messages, azure
from mirascope.integrations.otel import configure, with_otel

configure()


@with_otel()
@azure.call("gpt-4o-mini")
def recommend_book(genre: str) -> Messages.Type:
    return Messages.User(f"Recommend a {genre} book")


print(recommend_book("fantasy"))



from typing import Annotated

from mirascope.core import groq
from pydantic import AfterValidator, ValidationError


def is_upper(v: str) -> str:
    assert v.isupper(), "Must be uppercase"
    return v


def get_book_author(title: Annotated[str, AfterValidator(is_upper)]) -> str:
    """Returns the author of the book with the given title

    Args:
        title: The title of the book.
    """
    if title == "THE NAME OF THE WIND":
        return "Patrick Rothfuss"
    elif title == "MISTBORN: THE FINAL EMPIRE":
        return "Brandon Sanderson"
    else:
        return "Unknown"


@groq.call("llama-3.1-70b-versatile", tools=[get_book_author])
def identify_author(book: str) -> str:
    return f"Who wrote {book}?"


response = identify_author("The Name of the Wind")
try:
    if tool := response.tool:
        print(tool.call())
    else:
        print(response.content)
except ValidationError as e:
    print(e)
    # > 1 validation error for GetBookAuthor
    #   title
    #     Assertion failed, Must be uppercase [type=assertion_error, input_value='The Name of the Wind', input_type=str]
    #       For further information visit https://errors.pydantic.dev/2.8/v/assertion_error



from mirascope.core import groq


@groq.call("llama-3.1-70b-versatile", response_model=list[str])
def extract_book(texts: list[str]) -> str:
    return f"Extract book titles from {texts}"


book = extract_book(
    [
        "The Name of the Wind by Patrick Rothfuss",
        "Mistborn: The Final Empire by Brandon Sanderson",
    ]
)
print(book)
# Output: ["The Name of the Wind", "Mistborn: The Final Empire"]



from mirascope.core import Messages, gemini


@gemini.call("gemini-1.5-flash")
def recommend_book(genre: str) -> Messages.Type:
    return Messages.User(f"Recommend a {genre} book")


try:
    response = recommend_book("fantasy")
    print(response.content)
except Exception as e:
    print(f"Error: {str(e)}")



from anthropic import AsyncAnthropic
from mirascope.core import anthropic, prompt_template


@anthropic.call("claude-3-5-sonnet-20240620", client=AsyncAnthropic())
@prompt_template("Recommend a {genre} book")
async def recommend_book_async(genre: str): ...



"""The `create_factory` method for generating provider specific create decorators."""

import datetime
from collections.abc import Awaitable, Callable, Coroutine
from functools import wraps
from typing import Any, ParamSpec, TypeVar, cast, overload

from ._utils import (
    SameSyncAndAsyncClientSetupCall,
    SetupCall,
    fn_is_async,
    get_dynamic_configuration,
    get_fn_args,
    get_metadata,
    get_possible_user_message_param,
    is_prompt_template,
)
from .call_params import BaseCallParams
from .call_response import BaseCallResponse
from .dynamic_config import BaseDynamicConfig
from .messages import Messages
from .prompt import prompt_template
from .tool import BaseTool

_BaseCallResponseT = TypeVar("_BaseCallResponseT", bound=BaseCallResponse)
_SameSyncAndAsyncClientT = TypeVar("_SameSyncAndAsyncClientT", contravariant=True)
_SyncBaseClientT = TypeVar("_SyncBaseClientT", contravariant=True)
_AsyncBaseClientT = TypeVar("_AsyncBaseClientT", contravariant=True)
_BaseDynamicConfigT = TypeVar("_BaseDynamicConfigT", bound=BaseDynamicConfig)
_ParsedOutputT = TypeVar("_ParsedOutputT")
_BaseCallParamsT = TypeVar("_BaseCallParamsT", bound=BaseCallParams)
_ResponseT = TypeVar("_ResponseT")
_ResponseChunkT = TypeVar("_ResponseChunkT")
_AsyncResponseT = TypeVar("_AsyncResponseT")
_AsyncResponseChunkT = TypeVar("_AsyncResponseChunkT")
_BaseToolT = TypeVar("_BaseToolT", bound=BaseTool)
_P = ParamSpec("_P")


def create_factory(  # noqa: ANN202
    *,
    TCallResponse: type[_BaseCallResponseT],
    setup_call: SameSyncAndAsyncClientSetupCall[
        _SameSyncAndAsyncClientT,
        _BaseDynamicConfigT,
        _BaseCallParamsT,
        _ResponseT,
        _ResponseChunkT,
        _AsyncResponseT,
        _AsyncResponseChunkT,
        _BaseToolT,
    ]
    | SetupCall[
        _SyncBaseClientT,
        _AsyncBaseClientT,
        _BaseDynamicConfigT,
        _BaseCallParamsT,
        _ResponseT,
        _ResponseChunkT,
        _AsyncResponseT,
        _AsyncResponseChunkT,
        _BaseToolT,
    ],
):
    """Returns the wrapped function with the provider specific interfaces."""

    @overload
    def decorator(
        fn: Callable[_P, _BaseDynamicConfigT],
        model: str,
        tools: list[type[BaseTool] | Callable] | None,
        output_parser: Callable[[_BaseCallResponseT], _ParsedOutputT] | None,
        json_mode: bool,
        client: _SameSyncAndAsyncClientT | _SyncBaseClientT | None,
        call_params: _BaseCallParamsT,
    ) -> Callable[_P, _BaseCallResponseT | _ParsedOutputT]: ...

    @overload
    def decorator(
        fn: Callable[_P, Messages.Type],
        model: str,
        tools: list[type[BaseTool] | Callable] | None,
        output_parser: Callable[[_BaseCallResponseT], _ParsedOutputT] | None,
        json_mode: bool,
        client: _SameSyncAndAsyncClientT | _SyncBaseClientT | None,
        call_params: _BaseCallParamsT,
    ) -> Callable[_P, _BaseCallResponseT | _ParsedOutputT]: ...

    @overload
    def decorator(
        fn: Callable[
            _P,
            Awaitable[_BaseDynamicConfigT] | Coroutine[Any, Any, _BaseDynamicConfigT],
        ],
        model: str,
        tools: list[type[BaseTool] | Callable] | None,
        output_parser: Callable[[_BaseCallResponseT], _ParsedOutputT] | None,
        json_mode: bool,
        client: _SameSyncAndAsyncClientT | _AsyncBaseClientT | None,
        call_params: _BaseCallParamsT,
    ) -> Callable[
        _P,
        Awaitable[_BaseCallResponseT | _ParsedOutputT],
    ]: ...

    @overload
    def decorator(
        fn: Callable[_P, Awaitable[Messages.Type] | Coroutine[Any, Any, Messages.Type]],
        model: str,
        tools: list[type[BaseTool] | Callable] | None,
        output_parser: Callable[[_BaseCallResponseT], _ParsedOutputT] | None,
        json_mode: bool,
        client: _SameSyncAndAsyncClientT | _AsyncBaseClientT | None,
        call_params: _BaseCallParamsT,
    ) -> Callable[
        _P,
        Awaitable[_BaseCallResponseT | _ParsedOutputT],
    ]: ...

    def decorator(
        fn: Callable[_P, _BaseDynamicConfigT]
        | Callable[_P, Messages.Type]
        | Callable[
            _P,
            Awaitable[_BaseDynamicConfigT] | Coroutine[Any, Any, _BaseDynamicConfigT],
        ]
        | Callable[_P, Awaitable[Messages.Type] | Coroutine[Any, Any, Messages.Type]],
        model: str,
        tools: list[type[BaseTool] | Callable] | None,
        output_parser: Callable[[_BaseCallResponseT], _ParsedOutputT] | None,
        json_mode: bool,
        client: _SameSyncAndAsyncClientT | _AsyncBaseClientT | _SyncBaseClientT | None,
        call_params: _BaseCallParamsT,
    ) -> Callable[
        _P,
        _BaseCallResponseT
        | _ParsedOutputT
        | Awaitable[_BaseCallResponseT | _ParsedOutputT],
    ]:
        if not is_prompt_template(fn):
            fn = cast(
                Callable[_P, Messages.Type] | Callable[_P, Awaitable[Messages.Type]], fn
            )
            fn = prompt_template()(fn)
            fn = cast(
                Callable[_P, _BaseDynamicConfigT]
                | Callable[_P, Awaitable[_BaseDynamicConfigT]],
                fn,
            )
        fn._model = model  # pyright: ignore [reportFunctionMemberAccess]
        if fn_is_async(fn):

            @wraps(fn)
            async def inner_async(
                *args: _P.args, **kwargs: _P.kwargs
            ) -> TCallResponse | _ParsedOutputT:
                fn_args = get_fn_args(fn, args, kwargs)
                dynamic_config = await get_dynamic_configuration(fn, args, kwargs)
                create, prompt_template, messages, tool_types, call_kwargs = setup_call(  # pyright: ignore [reportCallIssue]
                    model=model,
                    client=client,  # pyright: ignore [reportArgumentType]
                    fn=fn,
                    fn_args=fn_args,
                    dynamic_config=dynamic_config,
                    tools=tools,
                    json_mode=json_mode,
                    call_params=call_params,
                    extract=False,
                )
                start_time = datetime.datetime.now().timestamp() * 1000
                response = await create(stream=False, **call_kwargs)
                end_time = datetime.datetime.now().timestamp() * 1000
                output = TCallResponse(
                    metadata=get_metadata(fn, dynamic_config),
                    response=response,
                    tool_types=tool_types,  # pyright: ignore [reportArgumentType]
                    prompt_template=prompt_template,
                    fn_args=fn_args,
                    dynamic_config=dynamic_config,
                    messages=messages,
                    call_params=call_params,
                    call_kwargs=call_kwargs,
                    user_message_param=get_possible_user_message_param(messages),
                    start_time=start_time,
                    end_time=end_time,
                )
                output._model = model
                return output if not output_parser else output_parser(output)

            return inner_async
        else:

            @wraps(fn)
            def inner(
                *args: _P.args, **kwargs: _P.kwargs
            ) -> TCallResponse | _ParsedOutputT:
                fn_args = get_fn_args(fn, args, kwargs)
                dynamic_config = get_dynamic_configuration(fn, args, kwargs)
                create, prompt_template, messages, tool_types, call_kwargs = setup_call(  # pyright: ignore [reportCallIssue]
                    model=model,
                    client=client,  # pyright: ignore [reportArgumentType]
                    fn=fn,
                    fn_args=fn_args,
                    dynamic_config=dynamic_config,
                    tools=tools,
                    json_mode=json_mode,
                    call_params=call_params,
                    extract=False,
                )
                start_time = datetime.datetime.now().timestamp() * 1000
                response = create(stream=False, **call_kwargs)
                end_time = datetime.datetime.now().timestamp() * 1000
                output = TCallResponse(
                    metadata=get_metadata(fn, dynamic_config),
                    response=response,
                    tool_types=tool_types,  # pyright: ignore [reportArgumentType]
                    prompt_template=prompt_template,
                    fn_args=fn_args,
                    dynamic_config=dynamic_config,
                    messages=messages,
                    call_params=call_params,
                    call_kwargs=call_kwargs,
                    user_message_param=get_possible_user_message_param(messages),
                    start_time=start_time,
                    end_time=end_time,
                )
                output._model = model
                return output if not output_parser else output_parser(output)

            return inner

    return decorator



"""The Mirascope Vertex Module."""

from typing import TypeAlias

from google.cloud.aiplatform_v1beta1.types import FunctionResponse
from vertexai.generative_models import Content

from ..base import BaseMessageParam
from ._call import vertex_call
from ._call import vertex_call as call
from .call_params import VertexCallParams
from .call_response import VertexCallResponse
from .call_response_chunk import VertexCallResponseChunk
from .dynamic_config import VertexDynamicConfig
from .stream import VertexStream
from .tool import VertexTool

VertexMessageParam: TypeAlias = Content | FunctionResponse | BaseMessageParam

__all__ = [
    "call",
    "VertexCallParams",
    "VertexCallResponse",
    "VertexCallResponseChunk",
    "VertexDynamicConfig",
    "VertexMessageParam",
    "VertexStream",
    "VertexTool",
    "vertex_call",
]



"""Tests the `bedrock._utils.convert_message_params` function."""

import pytest

from mirascope.core.base import (
    AudioPart,
    BaseMessageParam,
    ImagePart,
    TextPart,
)
from mirascope.core.bedrock import BedrockMessageParam
from mirascope.core.bedrock._utils._convert_message_params import (
    convert_message_params,
)


def test_convert_message_params() -> None:
    """Tests the `convert_message_params` function."""

    message_params: list[BaseMessageParam | BedrockMessageParam] = [
        {"role": "user", "content": [{"type": "text", "text": "Hello"}]},  # pyright: ignore [reportArgumentType, reportAssignmentType]
        BaseMessageParam(role="user", content="Hello"),
        BaseMessageParam(
            role="user",
            content=[
                TextPart(type="text", text="Hello"),
                ImagePart(
                    type="image", media_type="image/jpeg", image=b"image", detail="auto"
                ),
            ],
        ),
    ]
    converted_message_params = convert_message_params(message_params)
    assert converted_message_params == [
        {"content": [{"text": "Hello", "type": "text"}], "role": "user"},
        {"content": [{"text": "Hello"}], "role": "user"},
        {
            "content": [{"text": "Hello"}, {"bytes": b"image", "format": "image/jpeg"}],
            "role": "user",
        },
    ]

    with pytest.raises(
        ValueError,
        match="Unsupported image media type: image/svg. Bedrock currently only "
        "supports JPEG, PNG, GIF, and WebP images.",
    ):
        convert_message_params(
            [
                BaseMessageParam(
                    role="user",
                    content=[
                        ImagePart(
                            type="image",
                            media_type="image/svg",
                            image=b"image",
                            detail=None,
                        )
                    ],
                )
            ]
        )

    with pytest.raises(
        ValueError,
        match="Bedrock currently only supports text and image parts. Part provided: audio",
    ):
        convert_message_params(
            [
                BaseMessageParam(
                    role="user",
                    content=[
                        AudioPart(type="audio", media_type="audio/mp3", audio=b"audio")
                    ],
                )
            ]
        )



"""usage docs: learn/calls.md#provider-specific-parameters"""

from __future__ import annotations

from collections.abc import Sequence

from cohere.types import (
    ChatConnector,
    ChatDocument,
    ChatMessage,
    ChatRequestPromptTruncation,
    ToolResult,
)
from typing_extensions import NotRequired

from ..base import BaseCallParams


class CohereCallParams(BaseCallParams):
    """The parameters to use when calling the Cohere API.

    [Cohere API Reference](https://docs.cohere.com/reference/chat)

    Attributes:
        chat_history: ...
        connectors: ...
        conversation_id: ...
        documents: ...
        frequency_penalty: ...
        k: ...
        max_input_tokens: ...
        max_tokens: ...
        p: ...
        preamble: ...
        presence_penalty: ...
        prompt_truncation: ...
        raw_prompting: ...
        search_queries_only: ...
        seed: ...
        stop_sequences: ...
        temperature: ...
        tool_results: ...
    """

    chat_history: NotRequired[Sequence[ChatMessage] | None]
    connectors: NotRequired[Sequence[ChatConnector] | None]
    conversation_id: NotRequired[str | None]
    documents: NotRequired[Sequence[ChatDocument] | None]
    frequency_penalty: NotRequired[float | None]
    k: NotRequired[int | None]
    max_input_tokens: NotRequired[int | None]
    max_tokens: NotRequired[int | None]
    p: NotRequired[float | None]
    preamble: NotRequired[str | None]
    presence_penalty: NotRequired[float | None]
    prompt_truncation: NotRequired[ChatRequestPromptTruncation | None]
    raw_prompting: NotRequired[bool | None]
    search_queries_only: NotRequired[bool | None]
    seed: NotRequired[int | None]
    stop_sequences: NotRequired[Sequence[str] | None]
    temperature: NotRequired[float | None]
    tool_results: NotRequired[Sequence[ToolResult] | None]



from mirascope.core import anthropic
from mirascope.integrations.otel import with_hyperdx


@with_hyperdx()
@anthropic.call("claude-3-5-sonnet-20240620")
def recommend_book(genre: str) -> str:
    return f"Recommend a {genre} book"


print(recommend_book("fantasy"))



from mirascope.core import Messages, mistral
from mistralai.async_client import MistralAsyncClient


@mistral.call("mistral-large-latest", client=MistralAsyncClient())
async def recommend_book_async(genre: str) -> Messages.Type:
    return Messages.User(f"Recommend a {genre} book")



from litellm import completion


def recommend_book(genre: str) -> str:
    response = completion(
        model="gpt-4o-mini",
        messages=[{"role": "user", "content": f"Recommend a {genre} book"}],
    )
    return str(response.choices[0].message.content)  # type: ignore


output = recommend_book("fantasy")
print(output)



"""This module contains the type definition for the Anthropic call keyword arguments."""

from anthropic.types import MessageParam, ToolParam

from ..base import BaseCallKwargs
from .call_params import AnthropicCallParams


class AnthropicCallKwargs(AnthropicCallParams, BaseCallKwargs[ToolParam]):
    model: str
    messages: list[MessageParam]



from mirascope.core import groq


@groq.call("llama-3.1-70b-versatile")
def summarize(text: str) -> str:
    return f"Summarize this text: {text}"


@groq.call("llama-3.1-70b-versatile")
def summarize_and_translate(text: str, language: str) -> str:
    summary = summarize(text)
    return f"Translate this text to {language}: {summary.content}"


response = summarize_and_translate("Long English text here...", "french")
print(response.content)



from anthropic import Anthropic
from mirascope.core import Messages, anthropic


@anthropic.call("claude-3-5-sonnet-20240620", client=Anthropic())
def recommend_book(genre: str) -> Messages.Type:
    return Messages.User(f"Recommend a {genre} book")



"""Tests the `gemini._utils.calculate_cost` function."""

from mirascope.core.gemini._utils._calculate_cost import calculate_cost


def test_calculate_cost() -> None:
    """Tests the `calculate_cost` function."""
    assert calculate_cost(None, None, model="gemini-1.5-flash") is None
    assert calculate_cost(1, 1, model="unknown") is None



from mirascope.core import BaseMessageParam, azure


@azure.call("gpt-4o-mini", stream=True)
def recommend_book(genre: str) -> list[BaseMessageParam]:
    return [BaseMessageParam(role="user", content=f"Recommend a {genre} book")]


try:
    for chunk, _ in recommend_book("fantasy"):
        print(chunk.content, end="", flush=True)
except Exception as e:
    print(f"Error: {str(e)}")



from cohere import AsyncClient
from mirascope.core import BaseMessageParam, cohere


@cohere.call("command-r-plus", client=AsyncClient())
async def recommend_book_async(genre: str) -> list[BaseMessageParam]:
    return [BaseMessageParam(role="user", content=f"Recommend a {genre} book")]



import boto3

bedrock_client = boto3.client(service_name="bedrock-runtime")


def get_book_author(title: str) -> str:
    if title == "The Name of the Wind":
        return "Patrick Rothfuss"
    elif title == "Mistborn: The Final Empire":
        return "Brandon Sanderson"
    else:
        return "Unknown"


def identify_author(book: str) -> str:
    messages = [{"role": "user", "content": [{"text": f"Who wrote {book}?"}]}]
    tool_config = {
        "tools": [
            {
                "toolSpec": {
                    "name": "get_book_author",
                    "description": "Returns the author of the book with the given title.",
                    "inputSchema": {
                        "json": {
                            "type": "object",
                            "properties": {
                                "title": {
                                    "type": "string",
                                    "description": "The title of the book.",
                                }
                            },
                            "required": ["title"],
                        }
                    },
                }
            }
        ]
    }
    response = bedrock_client.converse(
        modelId="anthropic.claude-3-haiku-20240307-v1:0",
        messages=messages,
        toolConfig=tool_config,
    )
    content = ""
    output_message = response["output"]["message"]
    messages.append(output_message)
    stop_reason = response["stopReason"]

    if stop_reason == "tool_use":
        tool_requests = output_message["content"]
        for tool_request in tool_requests:
            if "toolUse" in tool_request:
                tool = tool_request["toolUse"]
                if tool["name"] == "get_book_author":
                    return get_book_author(**tool["input"])
    for content_piece in output_message["content"]:
        if "text" in content_piece:
            content += content_piece["text"]
    return content


author = identify_author("The Name of the Wind")
print(author)



from .calls import BaseCall
from .extractors import BaseExtractor, ExtractedType
from .prompts import tags
from .types import BaseCallParams

__all__ = [
    "BaseCall",
    "BaseCallParams",
    "BaseExtractor",
    "ExtractedType",
    "tags",
]



import asyncio

from mirascope.core import Messages, cohere


@cohere.call("command-r-plus")
async def recommend_book(genre: str) -> Messages.Type:
    return Messages.User(f"Recommend a {genre} book")


async def main():
    response = await recommend_book("fantasy")
    print(response.content)


asyncio.run(main())



from mirascope.core import bedrock
from botocore.exceptions import ClientError


@bedrock.call(model="anthropic.claude-3-haiku-20240307-v1:0", stream=True)
def recommend_book(genre: str) -> str:
    return f"Recommend a {genre} book"


try:
    for chunk, _ in recommend_book("fantasy"):
        print(chunk.content, end="", flush=True)
except ClientError as e:
    print(f"Error: {str(e)}")



"""Tests the `vertex._utils.calculate_cost` function."""

import pytest

from mirascope.core.vertex._utils._calculate_cost import calculate_cost


def test_calculate_cost() -> None:
    """Tests the `calculate_cost` function for Vertex AI Gemini models."""
    # Test None inputs
    assert calculate_cost(None, None, model="gemini-1.5-pro") is None

    # Test unknown model
    assert calculate_cost(1, 1, model="unknown") is None

    # Test normal cases with short context
    assert calculate_cost(
        1000, 1000, model="gemini-1.5-flash", context_length=100000
    ) == pytest.approx(0.00009375)
    assert calculate_cost(
        1000, 1000, model="gemini-1.5-pro", context_length=100000
    ) == pytest.approx(0.005)
    assert calculate_cost(
        1000, 1000, model="gemini-1.0-pro", context_length=100000
    ) == pytest.approx(0.0005)

    # Test long context cases
    assert calculate_cost(
        1000, 1000, model="gemini-1.5-flash", context_length=150000
    ) == pytest.approx(0.0001875)
    assert calculate_cost(
        1000, 1000, model="gemini-1.5-pro", context_length=150000
    ) == pytest.approx(0.01)

    # Test Gemini 1.0 Pro with long context (should return None)
    assert (
        calculate_cost(1000, 1000, model="gemini-1.0-pro", context_length=150000)
        is None
    )

    # Test zero input
    assert calculate_cost(0, 0, model="gemini-1.5-pro") == 0

    # Test very large input
    large_input = 1_000_000  # 1 million characters
    assert calculate_cost(
        large_input, large_input, model="gemini-1.5-pro"
    ) == pytest.approx(5)

    # Test fractional input
    assert calculate_cost(500.5, 500.5, model="gemini-1.5-pro") == pytest.approx(
        0.0025025
    )


def test_calculate_cost_edge_cases() -> None:
    """Tests edge cases for the `calculate_cost` function."""

    # Test exactly 128K context boundary
    assert calculate_cost(
        1000, 1000, model="gemini-1.5-pro", context_length=128000
    ) == pytest.approx(0.005)
    assert calculate_cost(
        1000, 1000, model="gemini-1.5-pro", context_length=128001
    ) == pytest.approx(0.01)



from mirascope.core import vertex
from vertexai.generative_models import GenerativeModel


@vertex.call("", client=GenerativeModel(model_name="gemini-1.5-flash"))
def recommend_book(genre: str) -> str:
    return f"Recommend a {genre} book"



import json

from mirascope.core import BaseMessageParam, azure


@azure.call("gpt-4o-mini", json_mode=True)
def get_book_info(book_title: str) -> list[BaseMessageParam]:
    return [
        BaseMessageParam(
            role="user", content=f"Provide the author and genre of {book_title}"
        )
    ]


response = get_book_info("The Name of the Wind")
print(json.loads(response.content))
# Output: {'author': 'Patrick Rothfuss', 'genre': 'Fantasy'}



from groq import GroqError
from mirascope.core import groq, prompt_template


@groq.call("llama-3.1-70b-versatile", stream=True)
@prompt_template("Recommend a {genre} book")
def recommend_book(genre: str): ...


try:
    for chunk, _ in recommend_book("fantasy"):
        print(chunk.content, end="", flush=True)
except GroqError as e:
    print(f"Error: {str(e)}")



from mirascope.core import bedrock, prompt_template


@bedrock.call("anthropic.claude-3-haiku-20240307-v1:0")
@prompt_template("Summarize this text: {text}")
def summarize(text: str): ...


@bedrock.call("anthropic.claude-3-haiku-20240307-v1:0")
@prompt_template("Translate this text to {language}: {text}")
def translate(text: str, language: str): ...


summary = summarize("Long English text here...")
translation = translate(summary.content, "french")
print(translation.content)



from groq import AsyncGroq
from mirascope.core import groq, prompt_template


@groq.call("llama-3.1-70b-versatile", client=AsyncGroq())
@prompt_template("Recommend a {genre} book")
async def recommend_book_async(genre: str): ...



from mirascope.core import gemini, prompt_template


def get_book_author(title: str) -> str:
    """Returns the author of the book with the given title

    Args:
        title: The title of the book.
    """
    if title == "The Name of the Wind":
        return "Patrick Rothfuss"
    elif title == "Mistborn: The Final Empire":
        return "Brandon Sanderson"
    else:
        return "Unknown"


@gemini.call("gemini-1.5-flash", tools=[get_book_author])
@prompt_template("Who wrote {books}?")
def identify_authors(books: list[str]): ...


response = identify_authors(["The Name of the Wind", "Mistborn: The Final Empire"])
if tools := response.tools:
    for tool in tools:
        print(tool.call())
else:
    print(response.content)



import logfire
from mirascope.core import Messages, mistral
from mirascope.integrations.logfire import with_logfire
from pydantic import BaseModel

logfire.configure()


class Book(BaseModel):
    title: str
    author: str


@with_logfire()
@mistral.call("mistral-large-latest", response_model=Book)
def recommend_book(genre: str) -> Messages.Type:
    return Messages.User(f"Recommend a {genre} book")


print(recommend_book("fantasy"))



"""Tests the `_utils.get_template_values` module."""

import pytest
from pydantic import BaseModel

from mirascope.core.base._utils._get_template_values import get_template_values


def test_get_template_values() -> None:
    """Test the get_template_values function."""

    class Book(BaseModel):
        title: str

    book = Book(title="book title")

    template_variables: list[tuple[str, str | None]] = [
        ("self", ""),
        ("self.var0", ""),
        ("book.title", ""),
        ("var1", ""),
        ("var2", ""),
        ("var3", "list"),
        ("var4", "list"),
        ("var5", "lists"),
        ("var6", "lists"),
    ]
    attrs = {
        "self": {"var0": "value0"},
        "book": book,
        "var1": "value1",
        "var2": ["value2", "value3"],
        "var3": ["value2", "value3"],
        "var4": [],
        "var5": [["value4", "value5"], ["value6", "value7"]],
        "var6": [],
    }
    expected = {
        "self": {"var0": "value0"},
        "book": book,
        "var1": "value1",
        "var2": ["value2", "value3"],
        "var3": "value2\nvalue3",
        "var4": "",
        "var5": "value4\nvalue5\n\nvalue6\nvalue7",
        "var6": "",
    }
    assert get_template_values(template_variables, attrs) == expected

    with pytest.raises(
        ValueError,
        match="Template variable 'var3' must be a list when using the 'list' format "
        "spec.",
    ):
        get_template_values([("var3", "list")], {"var3": "value"})

    list_of_lists_match = "Template variable 'var5' must be a list of lists when using "
    "the 'lists' format spec."

    with pytest.raises(ValueError, match=list_of_lists_match):
        get_template_values([("var5", "lists")], {"var5": "value"})

    with pytest.raises(ValueError, match=list_of_lists_match):
        get_template_values([("var5", "lists")], {"var5": ["value"]})



"""Get the JSON output from a completion response."""

import json

from ..call_response import BedrockCallResponse
from ..call_response_chunk import BedrockCallResponseChunk


def get_json_output(
    response: BedrockCallResponse | BedrockCallResponseChunk, json_mode: bool
) -> str:
    """Get the JSON output from a completion response."""
    if isinstance(response, BedrockCallResponse):
        if json_mode and response.content:
            return response.content
        elif message := response.message:
            tool_calls = [t for c in message["content"] if (t := c.get("toolUse"))]
            if tool_calls and (tool_call_input := tool_calls[0].get("input", {})):
                return json.dumps(tool_call_input)
        raise ValueError("No tool call or JSON object found in response.")
    else:
        if json_mode:
            return response.content
        elif (
            (content_block_delta := response.chunk.get("contentBlockDelta"))
            and (tool_use := content_block_delta["delta"].get("toolUse"))
            and (tool_use_input := tool_use.get("input"))
        ):
            return tool_use_input
    return ""



from mirascope.core import bedrock, prompt_template


@bedrock.call("anthropic.claude-3-haiku-20240307-v1:0", stream=True)
@prompt_template("Recommend a {genre} book")
def recommend_book(genre: str): ...


stream = recommend_book("fantasy")
for chunk, _ in stream:
    print(chunk.content, end="", flush=True)

print(f"Content: {stream.content}")

call_response = stream.construct_call_response()
print(f"Usage: {call_response.usage}")



"""This module contains utility functions for base types."""

import inspect
from enum import Enum
from typing import (
    Annotated,
    Any,
    Literal,
    TypeAlias,
    Union,
    get_origin,
)

from typing_extensions import TypeIs

BaseType: TypeAlias = str | int | float | bool | bytes | list | set | tuple | dict


def is_base_type(type_: Any) -> TypeIs[type[BaseType]]:  # noqa: ANN401
    """Check if a type is a base type."""
    base_types = {str, int, float, bool, bytes, list, set, tuple, dict}
    return (
        (inspect.isclass(type_) and issubclass(type_, Enum))
        or type_ in base_types
        or get_origin(type_) in base_types.union({Literal, Union, Annotated})
    )



from mirascope.core import BaseTool, Messages, litellm
from pydantic import Field


class GetBookAuthor(BaseTool):
    """Returns the author of the book with the given title."""

    title: str = Field(..., description="The title of the book.")

    def call(self) -> str:
        if self.title == "The Name of the Wind":
            return "Patrick Rothfuss"
        elif self.title == "Mistborn: The Final Empire":
            return "Brandon Sanderson"
        else:
            return "Unknown"


@litellm.call("gpt-4o-mini", tools=[GetBookAuthor])
def identify_author(book: str) -> Messages.Type:
    return Messages.User(f"Who wrote {book}?")


response = identify_author("The Name of the Wind")
if tool := response.tool:
    print(tool.call())
    # Output: Patrick Rothfuss
    print(f"Original tool call: {tool.tool_call}")
else:
    print(response.content)



from mirascope.core import vertex


@vertex.call(model="gemini-1.5-flash")
def recommend_book(genre: str) -> str:
    return f"Recommend a {genre} book"


try:
    response = recommend_book("fantasy")
    print(response.content)
except Exception as e:
    print(f"Error: {str(e)}")



from mirascope.core import groq


@groq.call("llama-3.1-70b-versatile")
def recommend_book(genre: str) -> str:
    return f"Recommend a {genre} book"


response = recommend_book("fantasy")
print(response.content)



from mirascope.core import Messages, anthropic
from mirascope.integrations.langfuse import with_langfuse


@with_langfuse()
@anthropic.call("claude-3-5-sonnet-20240620")
def recommend_book(genre: str) -> Messages.Type:
    return Messages.User(f"Recommend a {genre} book")


print(recommend_book("fantasy"))



from cohere import AsyncClient
from mirascope.core import Messages, cohere


@cohere.call("command-r-plus", client=AsyncClient())
async def recommend_book_async(genre: str) -> Messages.Type:
    return Messages.User(f"Recommend a {genre} book")



"""Tests the `call_response` module."""

from unittest.mock import MagicMock, patch

from mirascope.core.base.call_response import BaseCallResponse


def test_base_call_response() -> None:
    class MyCallResponse(BaseCallResponse):
        @property
        def content(self) -> str:
            return "content"

    patch.multiple(MyCallResponse, __abstractmethods__=set()).start()
    call_response = MyCallResponse(
        metadata={},
        response="",
        tool_types=None,
        prompt_template="",
        fn_args={},
        dynamic_config=None,
        messages=[],
        call_params={},
        call_kwargs={},
        user_message_param=None,
        start_time=0,
        end_time=0,
    )  # type: ignore
    assert str(call_response) == "content"
    tool = MagicMock()
    tool._name = lambda: "mock_tool"
    assert call_response.serialize_tool_types([tool], info=MagicMock()) == [
        {"type": "function", "name": "mock_tool"}
    ]



from mirascope.core import BaseMessageParam, vertex
from pydantic import BaseModel


class Librarian(BaseModel):
    history: list[vertex.VertexMessageParam] = []

    @vertex.call("gemini-1.5-flash")
    def _call(self, query: str) -> list[vertex.VertexMessageParam]:
        return [
            BaseMessageParam(role="system", content="You are a librarian"),
            *self.history,
            BaseMessageParam(role="user", content=query),
        ]

    def run(self) -> None:
        while True:
            query = input("(User): ")
            if query in ["exit", "quit"]:
                break
            print("(Assistant): ", end="", flush=True)
            response = self._call(query)
            print(response.content)
            self.history += [
                BaseMessageParam(role="user", content=query),
                response.message_param,
            ]


Librarian().run()



from collections.abc import AsyncGenerator, Callable, Generator
from typing import Any, ClassVar, Generic, TypeVar

from pydantic import model_validator
from tenacity import AsyncRetrying, Retrying
from typing_extensions import Self

from ...core.base import (
    BaseCallResponse,
    BaseCallResponseChunk,
    BasePrompt,
    BaseTool,
    _utils,
)
from ...core.base.dynamic_config import DynamicConfigFull
from .types import BaseCallParams
from .utils import retry_decorator

_BaseCallResponseT = TypeVar("_BaseCallResponseT", bound=BaseCallResponse)
_BaseCallResponseChunkT = TypeVar(
    "_BaseCallResponseChunkT", bound=BaseCallResponseChunk
)
_BaseToolT = TypeVar("_BaseToolT", bound=BaseTool)


class BaseCall(
    BasePrompt, Generic[_BaseCallResponseT, _BaseCallResponseChunkT, _BaseToolT]
):
    tags: ClassVar[list[str]] = []
    prompt_template: ClassVar[str] = ""

    call_params: ClassVar[BaseCallParams] = BaseCallParams(model="")

    _decorator: ClassVar[Callable] = lambda self: None
    _provider: ClassVar[str] = "NO PROVIDER"

    @model_validator(mode="after")
    def set_prompt_template(self) -> Self:
        self._prompt_template = self.prompt_template
        return self

    def dynamic_config(self) -> DynamicConfigFull:
        """Returns the dynamic configuration set from user provided information."""
        config: DynamicConfigFull = {
            "call_params": self.call_params.model_dump(exclude={"model", "tools"}),
            "tools": [
                tool
                if isinstance(tool, type)
                else _utils.convert_function_to_base_tool(tool, BaseTool)
                for tool in self.call_params.tools or []
            ],
        }
        if hasattr(self, "messages"):
            config["messages"] = self.messages()  # pyright: ignore [reportAttributeAccessIssue]
        return config

    def dump(
        self,
    ) -> dict[str, Any]:
        """Dumps the contents of the prompt into a dictionary."""
        return {
            "tags": self.tags,
            "template": _utils.get_prompt_template(self),
            "inputs": self.model_dump(),
        }

    def call(
        self,
        retries: int | Retrying = 0,
        client: Any | None = None,  # noqa: ANN401
        **kwargs: Any,  # noqa: ANN401
    ) -> _BaseCallResponseT:
        """A call to an LLM."""
        return self.run(
            self.__class__._decorator(self.call_params.model, client=client),
            retry_decorator(retries),
        )  # pyright: ignore [reportReturnType]

    async def call_async(
        self,
        retries: int | AsyncRetrying = 0,
        client: Any | None = None,  # noqa: ANN401
        **kwargs: Any,  # noqa: ANN401
    ) -> _BaseCallResponseT:
        """An asynchronous call to an LLM."""
        return await self.run_async(
            self.__class__._decorator(self.call_params.model, client=client),
            retry_decorator(retries),
        )  # pyright: ignore [reportReturnType]

    def stream(
        self,
        retries: int | Retrying = 0,
        client: Any | None = None,  # noqa: ANN401
        **kwargs: Any,  # noqa: ANN401
    ) -> Generator[tuple[_BaseCallResponseChunkT, _BaseToolT], None, None]:
        """A call to an LLM that streams the response in chunks."""
        yield from self.run(
            self.__class__._decorator(
                self.call_params.model, stream=True, client=client
            ),
            retry_decorator(retries),
        )  # pyright: ignore [reportReturnType]

    async def stream_async(
        self,
        retries: int | AsyncRetrying = 0,
        client: Any | None = None,  # noqa: ANN401
        **kwargs: Any,  # noqa: ANN401
    ) -> AsyncGenerator[tuple[_BaseCallResponseChunkT, _BaseToolT], None]:
        """A asynchronous call to an LLM that streams the response in chunks."""
        async for chunk, tool in self.run_async(  # pyright: ignore [reportGeneralTypeIssues]
            self.__class__._decorator(
                self.call_params.model, stream=True, client=client
            ),
            retry_decorator(retries),
        ):
            yield chunk, tool



from groq import GroqError
from mirascope.core import groq


@groq.call(model="llama-3.1-70b-versatile")
def recommend_book(genre: str) -> str:
    return f"Recommend a {genre} book"


try:
    response = recommend_book("fantasy")
    print(response.content)
except GroqError as e:
    print(f"Error: {str(e)}")



from mistralai.client import MistralClient

client = MistralClient()


def recommend_book(genre: str) -> str:
    completion = client.chat(
        model="mistral-large-latest",
        messages=[{"role": "user", "content": f"Recommend a {genre} book"}],
    )
    return completion.choices[0].message.content


output = recommend_book("fantasy")
print(output)



from typing import Annotated

from mirascope.core import BaseTool, azure, prompt_template
from pydantic import AfterValidator, Field, ValidationError


def is_upper(v: str) -> str:
    assert v.isupper(), "Must be uppercase"
    return v


class GetBookAuthor(BaseTool):
    """Returns the author of the book with the given title."""

    title: Annotated[str, AfterValidator(is_upper)] = Field(
        ..., description="The title of the book."
    )

    def call(self) -> str:
        if self.title == "THE NAME OF THE WIND":
            return "Patrick Rothfuss"
        elif self.title == "MISTBORN: THE FINAL EMPIRE":
            return "Brandon Sanderson"
        else:
            return "Unknown"


@azure.call("gpt-4o-mini", tools=[GetBookAuthor])
@prompt_template("Who wrote {book}?")
def identify_author(book: str): ...


response = identify_author("The Name of the Wind")
try:
    if tool := response.tool:
        print(tool.call())
    else:
        print(response.content)
except ValidationError as e:
    print(e)
    # > 1 validation error for GetBookAuthor
    #   title
    #     Assertion failed, Must be uppercase [type=assertion_error, input_value='The Name of the Wind', input_type=str]
    #       For further information visit https://errors.pydantic.dev/2.8/v/assertion_error



from mirascope.core import BaseMessageParam, bedrock
from mirascope.integrations.otel import configure, with_otel

configure()


@with_otel()
@bedrock.call("anthropic.claude-3-haiku-20240307-v1:0")
def recommend_book(genre: str) -> list[BaseMessageParam]:
    return [BaseMessageParam(role="user", content=f"Recommend a {genre} book.")]


print(recommend_book("fantasy"))



"""Tests the `azure._utils.get_json_output` module."""

from datetime import datetime

import pytest
from azure.ai.inference.models import (
    ChatChoice,
    ChatCompletions,
    ChatCompletionsToolCall,
    ChatResponseMessage,
    CompletionsUsage,
    FunctionCall,
    StreamingChatChoiceUpdate,
    StreamingChatCompletionsUpdate,
    StreamingChatResponseMessageUpdate,
    StreamingChatResponseToolCallUpdate,
)

from mirascope.core.azure._utils._get_json_output import get_json_output
from mirascope.core.azure.call_response import AzureCallResponse
from mirascope.core.azure.call_response_chunk import AzureCallResponseChunk


def test_get_json_output_call_response() -> None:
    """Tests the `get_json_output` function with a call response."""
    tool_call = ChatCompletionsToolCall(
        id="id",
        function=FunctionCall(
            arguments='{"title": "The Name of the Wind", "author": "Patrick Rothfuss"}',
            name="FormatBook",
        ),
    )
    choices = [
        ChatChoice(
            finish_reason="stop",
            index=0,
            message=ChatResponseMessage(
                content="json_output", role="assistant", tool_calls=[tool_call]
            ),
        )
    ]
    completion = ChatCompletions(
        id="id",
        choices=choices,
        created=datetime.fromtimestamp(0),
        model="gpt-4o",
        usage=CompletionsUsage(completion_tokens=1, prompt_tokens=1, total_tokens=2),
    )
    call_response = AzureCallResponse(
        metadata={},
        response=completion,
        tool_types=None,
        prompt_template="",
        fn_args={},
        dynamic_config=None,
        messages=[],
        call_params={},
        call_kwargs={},
        user_message_param=None,
        start_time=0,
        end_time=0,
    )
    assert get_json_output(call_response, json_mode=True) == "json_output"
    assert (
        get_json_output(call_response, json_mode=False)
        == '{"title": "The Name of the Wind", "author": "Patrick Rothfuss"}'
    )

    completion.choices[0].message.content = ""
    completion.choices[0].message.tool_calls = []
    with pytest.raises(
        ValueError, match="No tool call or JSON object found in response."
    ):
        get_json_output(call_response, json_mode=False)


def test_get_json_output_call_response_chunk() -> None:
    """Tests the `get_json_output` function with a call response chunk."""
    tool_call = StreamingChatResponseToolCallUpdate(
        id="id",
        function=FunctionCall(
            arguments='{"title": "The Name of the Wind", "author": "Patrick Rothfuss"}',
            name="FormatBook",
        ),
    )
    choices = [
        StreamingChatChoiceUpdate(
            delta=StreamingChatResponseMessageUpdate(
                content="json_output", tool_calls=[tool_call]
            ),
            index=0,
            finish_reason="stop",
        )
    ]
    chunk = StreamingChatCompletionsUpdate(
        id="id",
        choices=choices,
        created=datetime.fromtimestamp(0),
        model="gpt-4o",
        usage=CompletionsUsage(completion_tokens=1, prompt_tokens=1, total_tokens=2),
    )
    call_response_chunk = AzureCallResponseChunk(chunk=chunk)
    assert get_json_output(call_response_chunk, json_mode=True) == "json_output"
    assert (
        get_json_output(call_response_chunk, json_mode=False)
        == '{"title": "The Name of the Wind", "author": "Patrick Rothfuss"}'
    )

    chunk.choices[0].delta.tool_calls = None
    assert get_json_output(call_response_chunk, json_mode=False) == ""



from mirascope.core import openai
from pydantic import BaseModel

from fastapi import FastAPI

app = FastAPI()


class Book(BaseModel):
    title: str
    author: str


@app.post("/recommend_book")
@openai.call(model="gpt-4o-mini", response_model=Book)
def recommend_book(genre: str) -> str:
    return f"Recommend a {genre} book."


@app.post("/recommend_book_async")
@openai.call(model="gpt-4o-mini", response_model=Book)
async def recommend_book_async(genre: str) -> str:
    return f"Recommend a {genre} book."



"""This module contains the type definition for the Bedrock call keyword arguments."""

from ..base import BaseCallKwargs
from ._types import InternalBedrockMessageParam, ToolTypeDef
from .call_params import BedrockCallParams


class BedrockCallKwargs(BedrockCallParams, BaseCallKwargs[ToolTypeDef]):
    modelId: str
    messages: list[InternalBedrockMessageParam]



from unittest.mock import AsyncMock, MagicMock, patch

import pytest
from types_aiobotocore_bedrock_runtime import (
    BedrockRuntimeClient as AsyncBedrockRuntimeClient,
)

from mirascope.core.bedrock._utils._setup_call import (
    _extract_async_stream_fn,
    _extract_sync_stream_fn,
    _get_async_client,
    setup_call,
)
from mirascope.core.bedrock.tool import BedrockTool


@pytest.fixture()
def mock_base_setup_call() -> MagicMock:
    mock_setup_call = MagicMock()
    mock_setup_call.return_value = [MagicMock() for _ in range(3)] + [{}]
    return mock_setup_call


@pytest.mark.asyncio
async def test_get_async_client():
    mock_session = MagicMock()

    mock_client = MagicMock(spec=AsyncBedrockRuntimeClient)
    mock_session.create_client.return_value.__aenter__.return_value = mock_client

    client = await _get_async_client(mock_session)

    assert isinstance(client, MagicMock)
    assert client == mock_client
    mock_session.create_client.assert_called_once_with("bedrock-runtime")


def test_extract_sync_stream_fn():
    mock_fn = MagicMock()
    mock_fn.return_value = {
        "ResponseMetadata": {"RequestId": "123"},
        "stream": [
            {
                "contentBlockDelta": {"delta": [{"text": "Hello"}]},
            },
            {
                "contentBlockDelta": {"delta": [{"text": "World"}]},
            },
        ],
    }

    extracted_fn = _extract_sync_stream_fn(mock_fn, "test-model")
    result = list(extracted_fn())

    assert len(result) == 2
    assert result[0]["contentBlockDelta"]["delta"][0]["text"] == "Hello"  # pyright: ignore [reportTypedDictNotRequiredAccess, reportGeneralTypeIssues]
    assert result[1]["contentBlockDelta"]["delta"][0]["text"] == "World"  # pyright: ignore [reportTypedDictNotRequiredAccess, reportGeneralTypeIssues]
    assert all(chunk["responseMetadata"] == {"RequestId": "123"} for chunk in result)
    assert all(chunk["model"] == "test-model" for chunk in result)


@pytest.mark.asyncio
async def test_extract_async_stream_fn():
    async def async_generator():
        yield {"chunk1": "async_data1"}
        yield {"chunk2": "async_data2"}

    mock_fn = AsyncMock()
    mock_fn.return_value = {
        "ResponseMetadata": {"RequestId": "456"},
        "stream": async_generator(),
    }

    extracted_fn = _extract_async_stream_fn(mock_fn, "test-async-model")
    result = [chunk async for chunk in extracted_fn()]

    assert len(result) == 2
    assert result[0]["chunk1"] == "async_data1"  # pyright: ignore [reportGeneralTypeIssues]
    assert result[1]["chunk2"] == "async_data2"  # pyright: ignore [reportGeneralTypeIssues]
    assert all(chunk["responseMetadata"] == {"RequestId": "456"} for chunk in result)
    assert all(chunk["model"] == "test-async-model" for chunk in result)


@patch(
    "mirascope.core.bedrock._utils._setup_call.convert_message_params",
    new_callable=MagicMock,
)
@patch("mirascope.core.bedrock._utils._setup_call._utils", new_callable=MagicMock)
def test_setup_call(
    mock_utils: MagicMock,
    mock_convert_message_params: MagicMock,
    mock_base_setup_call: MagicMock,
) -> None:
    mock_utils.setup_call = mock_base_setup_call
    mock_base_setup_call.return_value[3] = {}
    fn = MagicMock()
    create, prompt_template, messages, tool_types, call_kwargs = setup_call(
        model="anthropic.claude-v2",
        client=None,
        fn=fn,
        fn_args={},
        dynamic_config=None,
        tools=None,
        json_mode=False,
        call_params={},
        extract=False,
    )
    assert prompt_template == mock_base_setup_call.return_value[0]
    assert tool_types == mock_base_setup_call.return_value[2]
    assert "modelId" in call_kwargs and call_kwargs["modelId"] == "anthropic.claude-v2"
    assert "messages" in call_kwargs and call_kwargs["messages"] == messages
    mock_base_setup_call.assert_called_once_with(fn, {}, None, None, BedrockTool, {})
    mock_convert_message_params.assert_called_once_with(
        mock_base_setup_call.return_value[1]
    )
    assert messages == mock_convert_message_params.return_value


@patch("mirascope.core.bedrock._utils._setup_call._utils", new_callable=MagicMock)
def test_setup_call_system_message(
    mock_utils: MagicMock, mock_base_setup_call: MagicMock
) -> None:
    mock_base_setup_call.return_value[1] = [
        {"role": "system", "content": [{"text": "system test"}]},
        {"role": "user", "content": [{"text": "user test"}]},
    ]
    mock_utils.setup_call = mock_base_setup_call
    mock_base_setup_call.return_value[3] = {}
    _, _, messages, _, _ = setup_call(
        model="anthropic.claude-v2",
        client=None,
        fn=MagicMock(),
        fn_args={},
        dynamic_config=None,
        tools=None,
        json_mode=False,
        call_params={},
        extract=False,
    )
    assert messages[0]["role"] == "user"
    assert messages[0]["content"] == [{"text": "user test"}]


@patch(
    "mirascope.core.bedrock._utils._setup_call.convert_message_params",
    new_callable=MagicMock,
)
@patch("mirascope.core.bedrock._utils._setup_call._utils", new_callable=MagicMock)
def test_setup_call_json_mode(
    mock_utils: MagicMock,
    mock_convert_message_params: MagicMock,
    mock_base_setup_call: MagicMock,
) -> None:
    mock_utils.setup_call = mock_base_setup_call
    mock_utils.json_mode_content = MagicMock(return_value="json_content")
    mock_base_setup_call.return_value[1] = [
        {"role": "user", "content": [{"text": "test"}]}
    ]
    mock_base_setup_call.return_value[3] = {"tools": MagicMock()}
    mock_convert_message_params.side_effect = lambda x: x
    _, _, messages, _, call_kwargs = setup_call(
        model="anthropic.claude-v2",
        client=None,
        fn=MagicMock(),
        fn_args={},
        dynamic_config=None,
        tools=None,
        json_mode=True,
        call_params={},
        extract=False,
    )
    assert messages[-1]["content"] == [{"text": "testjson_content"}]
    assert "tools" not in call_kwargs


@patch(
    "mirascope.core.bedrock._utils._setup_call.convert_message_params",
    new_callable=MagicMock,
)
@patch("mirascope.core.bedrock._utils._setup_call._utils", new_callable=MagicMock)
def test_setup_call_json_mode_no_text(
    mock_utils: MagicMock,
    mock_convert_message_params: MagicMock,
    mock_base_setup_call: MagicMock,
) -> None:
    mock_utils.setup_call = mock_base_setup_call
    mock_utils.json_mode_content = MagicMock(return_value="json_content")
    mock_base_setup_call.return_value[1] = [
        {"role": "user", "content": [{"image": "test_image"}]}
    ]
    mock_base_setup_call.return_value[3] = {"tools": MagicMock()}
    mock_convert_message_params.side_effect = lambda x: x
    _, _, messages, _, call_kwargs = setup_call(
        model="anthropic.claude-v2",
        client=None,
        fn=MagicMock(),
        fn_args={},
        dynamic_config=None,
        tools=None,
        json_mode=True,
        call_params={},
        extract=False,
    )
    assert messages[-1]["content"] == [
        {"image": "test_image"},
        {"text": "json_content"},
    ]
    assert "tools" not in call_kwargs


@patch(
    "mirascope.core.bedrock._utils._setup_call.convert_message_params",
    new_callable=MagicMock,
)
@patch("mirascope.core.bedrock._utils._setup_call._utils", new_callable=MagicMock)
def test_setup_call_extract(
    mock_utils: MagicMock,
    mock_convert_message_params: MagicMock,
    mock_base_setup_call: MagicMock,
) -> None:
    tool_mock = MagicMock(spec=BedrockTool, __name__="tool")
    tool_mock._name.return_value = "mock_tool"
    mock_base_setup_call.return_value[2] = [tool_mock]
    mock_base_setup_call.return_value[3] = {"toolConfig": {}}
    mock_utils.setup_call = mock_base_setup_call
    _, _, _, tool_types, call_kwargs = setup_call(
        model="anthropic.claude-v2",
        client=None,
        fn=MagicMock(),
        fn_args={},
        dynamic_config=None,
        tools=None,
        json_mode=False,
        call_params={},
        extract=True,
    )
    assert isinstance(tool_types, list)
    assert "toolConfig" in call_kwargs and call_kwargs["toolConfig"] == {
        "toolChoice": {"name": "mock_tool", "type": "tool"}
    }


@patch(
    "mirascope.core.bedrock._utils._setup_call.convert_message_params",
    new_callable=MagicMock,
)
@patch("mirascope.core.bedrock._utils._setup_call._utils", new_callable=MagicMock)
def test_setup_call_with_tools(
    mock_utils: MagicMock,
    mock_convert_message_params: MagicMock,
    mock_base_setup_call: MagicMock,
) -> None:
    mock_utils.setup_call = mock_base_setup_call
    mock_base_setup_call.return_value[3] = {"tools": [{"name": "test_tool"}]}
    fn = MagicMock()
    _, _, _, _, call_kwargs = setup_call(
        model="anthropic.claude-v2",
        client=None,
        fn=fn,
        fn_args={},
        dynamic_config=None,
        tools=None,
        json_mode=False,
        call_params={},
        extract=False,
    )
    assert "toolConfig" in call_kwargs
    assert call_kwargs["toolConfig"] == {"tools": [{"name": "test_tool"}]}


@patch("mirascope.core.bedrock._utils._setup_call._get_async_client")
@patch("mirascope.core.bedrock._utils._setup_call._utils", new_callable=MagicMock)
def test_setup_call_client_creation(
    mock_utils: MagicMock,
    mock_get_async_client: AsyncMock,
    mock_base_setup_call: MagicMock,
) -> None:
    mock_utils.setup_call = mock_base_setup_call
    mock_base_setup_call.return_value[1] = [
        {"role": "user", "content": [{"text": "user test"}]},
    ]
    mock_base_setup_call.return_value[3] = {}

    # Test sync client creation
    setup_call(
        model="anthropic.claude-v2",
        client=None,
        fn=MagicMock(),
        fn_args={},
        dynamic_config=None,
        tools=None,
        json_mode=False,
        call_params={},
        extract=False,
    )

    # Test async client creation
    async def async_fn(): ...

    setup_call(
        model="anthropic.claude-v2",
        client=None,
        fn=async_fn,
        fn_args={},
        dynamic_config=None,
        tools=None,
        json_mode=False,
        call_params={},
        extract=False,
    )
    mock_get_async_client.assert_called_once()

    # Test when client is provided
    mock_client = MagicMock()
    setup_call(
        model="anthropic.claude-v2",
        client=mock_client,
        fn=MagicMock(),
        fn_args={},
        dynamic_config=None,
        tools=None,
        json_mode=False,
        call_params={},
        extract=False,
    )
    assert mock_get_async_client.call_count == 1



from mirascope.core import BaseMessageParam, openai
from openai import OpenAI


@openai.call("gpt-4o-mini", client=OpenAI())
def recommend_book(genre: str) -> list[BaseMessageParam]:
    return [BaseMessageParam(role="user", content=f"Recommend a {genre} book")]



import json

from mirascope.core import anthropic


def only_json(response: anthropic.AnthropicCallResponse) -> str:
    json_start = response.content.index("{")
    json_end = response.content.rfind("}")
    return response.content[json_start : json_end + 1]


@anthropic.call("claude-3-5-sonnet-20240620", json_mode=True, output_parser=only_json)
def json_extraction(text: str, fields: list[str]) -> str:
    return f"Extract {fields} from the following text: {text}"


json_response = json_extraction(
    text="The capital of France is Paris",
    fields=["capital", "country"],
)
print(json.loads(json_response))



from typing import Annotated

from mirascope.core import BaseMessageParam, FromCallArgs, groq
from pydantic import BaseModel, model_validator
from typing_extensions import Self


class Book(BaseModel):
    title: str
    author: str


class Books(BaseModel):
    texts: Annotated[list[str], FromCallArgs()]
    books: list[Book]

    @model_validator(mode="after")
    def validate_output_length(self) -> Self:
        if len(self.texts) != len(self.books):
            raise ValueError("length mismatch...")
        return self


@groq.call("llama-3.1-70b-versatile", response_model=Books)
def extract_books(texts: list[str]) -> list[BaseMessageParam]:
    return [
        BaseMessageParam(
            role="user", content=f"Extract the books from these texts: {texts}"
        )
    ]


texts = [
    "The Name of the Wind by Patrick Rothfuss",
    "Mistborn: The Final Empire by Brandon Sanderson",
]
print(extract_books(texts))
# Output: texts=['The Name of the Wind by Patrick Rothfuss', 'Mistborn: The Final Empire by Brandon Sanderson'] books=[Book(title='The Name of the Wind', author='Patrick Rothfuss'), Book(title='Mistborn: The Final Empire', author='Brandon Sanderson')]



from mirascope.core import groq, prompt_template


@groq.call("llama-3.1-70b-versatile", call_params={"max_tokens": 512})
@prompt_template("Recommend a {genre} book")
def recommend_book(genre: str): ...


response = recommend_book("fantasy")
print(response.content)



import asyncio
from io import BytesIO

from typing import AsyncGenerator, Any

from pydub import AudioSegment
from pydub.playback import play

from mirascope.beta.openai import Realtime, record_as_stream, async_input


app = Realtime(
    "gpt-4o-realtime-preview-2024-10-01",
    turn_detection=None,
)


@app.receiver("audio")
async def receive_audio(response: AudioSegment, context: dict[str, Any]) -> None:
    play(response)


@app.receiver("audio_transcript")
async def receive_audio_transcript(response: str, context: dict[str, Any]) -> None:
    print(f"AI(audio_transcript): {response}")


@app.sender(wait_for_audio_transcript_response=True)
async def send_audio_as_stream(
    context: dict[str, Any],
) -> AsyncGenerator[BytesIO, None]:
    message = await async_input(
        "Press Enter to start recording or enter exit to shutdown app"
    )
    if message == "exit":
        raise asyncio.CancelledError

    async def wait_for_enter() -> str:
        return await async_input("Press Enter to stop recording...")

    async for stream in record_as_stream(custom_blocking_event=wait_for_enter):
        yield stream


asyncio.run(app.run())



from cohere import Client
from mirascope.core import cohere, prompt_template


@cohere.call("command-r-plus", client=Client())
@prompt_template("Recommend a {genre} book")
def recommend_book(genre: str): ...



import asyncio

from mirascope.core import bedrock


@bedrock.call(model="anthropic.claude-3-haiku-20240307-v1:0")
async def recommend_book(genre: str) -> str:
    return f"Recommend a {genre} book"


async def main():
    response = await recommend_book("fantasy")
    print(response.content)


asyncio.run(main())



from mirascope.core import BaseMessageParam, vertex
from tenacity import retry, stop_after_attempt, wait_exponential


@retry(
    stop=stop_after_attempt(3),
    wait=wait_exponential(multiplier=1, min=4, max=10),
)
@vertex.call("gemini-1.5-flash")
def recommend_book(genre: str) -> list[BaseMessageParam]:
    return [BaseMessageParam(role="user", content=f"Recommend a {genre} book")]


print(recommend_book("fantasy"))



from groq import Groq
from mirascope.core import Messages, groq


@groq.call("llama-3.1-70b-versatile", client=Groq())
def recommend_book(genre: str) -> Messages.Type:
    return Messages.User(f"Recommend a {genre} book")



"""The `call_factory` method for generating provider specific call decorators."""

from collections.abc import AsyncIterable, Callable, Iterable
from enum import Enum
from functools import partial
from typing import Annotated, TypeVar

from pydantic import BaseModel

from ._create import create_factory
from ._extract import extract_factory
from ._utils import (
    BaseType,
    GetJsonOutput,
    HandleStream,
    HandleStreamAsync,
    LLMFunctionDecorator,
    SameSyncAndAsyncClientSetupCall,
    SetupCall,
)
from ._utils._protocols import (
    AsyncLLMFunctionDecorator,
    CallDecorator,
    SyncLLMFunctionDecorator,
)
from .call_params import BaseCallParams
from .call_response import BaseCallResponse
from .call_response_chunk import BaseCallResponseChunk
from .dynamic_config import BaseDynamicConfig
from .stream import BaseStream, stream_factory
from .structured_stream import structured_stream_factory
from .tool import BaseTool

_BaseCallResponseT = TypeVar("_BaseCallResponseT", bound=BaseCallResponse)
_BaseCallResponseChunkT = TypeVar(
    "_BaseCallResponseChunkT", bound=BaseCallResponseChunk
)
_ResponseModelT = TypeVar(
    "_ResponseModelT", bound=BaseModel | BaseType | Enum | Annotated
)
_ParsedOutputT = TypeVar("_ParsedOutputT")
_BaseCallParamsT = TypeVar("_BaseCallParamsT", bound=BaseCallParams)
_BaseDynamicConfigT = TypeVar("_BaseDynamicConfigT", bound=BaseDynamicConfig)
_BaseStreamT = TypeVar("_BaseStreamT", bound=BaseStream)
_SameSyncAndAsyncClientT = TypeVar("_SameSyncAndAsyncClientT", contravariant=True)
_SyncBaseClientT = TypeVar("_SyncBaseClientT", contravariant=True)
_AsyncBaseClientT = TypeVar("_AsyncBaseClientT", contravariant=True)
_BaseToolT = TypeVar("_BaseToolT", bound=BaseTool)
_ResponseT = TypeVar("_ResponseT")
_ResponseChunkT = TypeVar("_ResponseChunkT")
_AsyncResponseT = TypeVar("_AsyncResponseT")
_AsyncResponseChunkT = TypeVar("_AsyncResponseChunkT")


def call_factory(  # noqa: ANN202
    *,
    TCallResponse: type[_BaseCallResponseT],
    TCallResponseChunk: type[_BaseCallResponseChunkT],
    TToolType: type[_BaseToolT],
    TStream: type[_BaseStreamT],
    default_call_params: _BaseCallParamsT,
    setup_call: SameSyncAndAsyncClientSetupCall[
        _SameSyncAndAsyncClientT,
        _BaseDynamicConfigT,
        _BaseCallParamsT,
        _ResponseT,
        _ResponseChunkT,
        _AsyncResponseT,
        _AsyncResponseChunkT,
        _BaseToolT,
    ]
    | SetupCall[
        _SyncBaseClientT,
        _AsyncBaseClientT,
        _BaseDynamicConfigT,
        _BaseCallParamsT,
        _ResponseT,
        _ResponseChunkT,
        _AsyncResponseT,
        _AsyncResponseChunkT,
        _BaseToolT,
    ],
    get_json_output: GetJsonOutput[_BaseCallResponseT | _BaseCallResponseChunkT],
    handle_stream: HandleStream[_ResponseChunkT, _BaseCallResponseChunkT, _BaseToolT],
    handle_stream_async: HandleStreamAsync[
        _AsyncResponseChunkT, _BaseCallResponseChunkT, _BaseToolT
    ],
) -> CallDecorator[
    _BaseCallResponseT,
    _BaseCallResponseChunkT,
    _BaseDynamicConfigT,
    _BaseCallParamsT,
    _BaseStreamT,
    _SyncBaseClientT,
    _AsyncBaseClientT,
    _SameSyncAndAsyncClientT,
]:
    """A factory method for creating provider-specific call decorators.

    Args:
        TCallResponse: The provider-specific `BaseCallResponse` type.
        TCallResponseChunk: The provider-specific `BaseCallResponseChunk` type.
        TToolType: The provider-specific `BaseTool` type.
        TStream: The provider-specific `BaseStream` type.
        default_call_params: The default call parameters to use, which must match the
            `TCallParams` type if provided.
        setup_call: The helper method for setting up a call, which returns the
            configured create function, the prompt template, the list of
            provider-specific messages, the list of provider-specific tool types, and
            the finalized `call_kwargs` with which to make the API call with the create
            function.
        get_json_output: The helper method for getting JSON output from a call response.
        handle_stream: The helper method for converting a provider's original stream
            generator into a generator that returns tuples of `(chunk, tool)` where
            `chunk` and `tool` are provider-specific `BaseCallResponseChunk` and
            `BaseTool` instances, respectively.
        handle_stream_async: The same helper method as `handle_stream` except for
            handling asynchronous streaming.
    """

    def base_call(
        model: str,
        *,
        stream: bool = False,
        tools: list[type[BaseTool] | Callable] | None = None,
        response_model: type[_ResponseModelT] | None = None,
        output_parser: Callable[[_BaseCallResponseT], _ParsedOutputT]
        | Callable[[_BaseCallResponseChunkT], _ParsedOutputT]
        | Callable[[_ResponseModelT], _ParsedOutputT]
        | None = None,
        json_mode: bool = False,
        client: _SameSyncAndAsyncClientT
        | _AsyncBaseClientT
        | _SyncBaseClientT
        | None = None,
        call_params: BaseCallParams | None = None,
    ) -> (
        AsyncLLMFunctionDecorator[
            _BaseDynamicConfigT,
            _BaseCallResponseT
            | _ParsedOutputT
            | BaseStream
            | _ResponseModelT
            | AsyncIterable[_ResponseModelT],
        ]
        | SyncLLMFunctionDecorator[
            _BaseDynamicConfigT,
            _BaseCallResponseT
            | _ParsedOutputT
            | BaseStream
            | _ResponseModelT
            | Iterable[_ResponseModelT],
        ]
        | LLMFunctionDecorator[
            _BaseDynamicConfigT,
            _BaseCallResponseT
            | _ParsedOutputT
            | BaseStream
            | _ResponseModelT
            | Iterable[_ResponseModelT],
            _BaseCallResponseT
            | _ParsedOutputT
            | BaseStream
            | _ResponseModelT
            | AsyncIterable[_ResponseModelT],
        ]
    ):
        if stream and output_parser:
            raise ValueError("Cannot use `output_parser` with `stream=True`.")

        if call_params is None:
            call_params = default_call_params

        if response_model:
            if stream:
                return partial(
                    structured_stream_factory(
                        TCallResponse=TCallResponse,
                        TCallResponseChunk=TCallResponseChunk,
                        TStream=TStream,
                        TToolType=TToolType,
                        setup_call=setup_call,
                        get_json_output=get_json_output,
                    ),
                    model=model,
                    response_model=response_model,
                    json_mode=json_mode,
                    client=client,
                    call_params=call_params,
                )  # pyright: ignore [reportReturnType, reportCallIssue]
            else:
                return partial(
                    extract_factory(
                        TCallResponse=TCallResponse,
                        TToolType=TToolType,
                        setup_call=setup_call,
                        get_json_output=get_json_output,
                    ),
                    model=model,
                    response_model=response_model,
                    output_parser=output_parser,
                    json_mode=json_mode,
                    client=client,
                    call_params=call_params,
                )  # pyright: ignore [reportCallIssue]

        if stream:
            return partial(
                stream_factory(
                    TCallResponse=TCallResponse,
                    TStream=TStream,
                    setup_call=setup_call,
                    handle_stream=handle_stream,
                    handle_stream_async=handle_stream_async,
                ),
                model=model,
                tools=tools,
                json_mode=json_mode,
                client=client,
                call_params=call_params,
            )  # pyright: ignore [reportReturnType, reportCallIssue]
        return partial(
            create_factory(TCallResponse=TCallResponse, setup_call=setup_call),
            model=model,
            tools=tools,
            output_parser=output_parser,
            json_mode=json_mode,
            client=client,
            call_params=call_params,
        )  # pyright: ignore [reportReturnType, reportCallIssue]

    return base_call  # pyright: ignore [reportReturnType]



from mirascope.core import BaseMessageParam, litellm


def get_book_author(title: str) -> str:
    """Returns the author of the book with the given title

    Example:
        {"title": "The Name of the Wind"}

    Args:
        title: The title of the book.
    """
    if title == "The Name of the Wind":
        return "Patrick Rothfuss"
    elif title == "Mistborn: The Final Empire":
        return "Brandon Sanderson"
    else:
        return "Unknown"


@litellm.call("gpt-4o-mini", tools=[get_book_author])
def identify_author(book: str) -> list[BaseMessageParam]:
    return [BaseMessageParam(role="user", content=f"Who wrote {book}?")]


response = identify_author("The Name of the Wind")
if tool := response.tool:
    print(tool.call())
else:
    print(response.content)



"""This module contains the type definition for the Mistral call keyword arguments."""

from typing import Any

from mistralai.models.chat_completion import ChatMessage

from ..base import BaseCallKwargs
from .call_params import MistralCallParams


class MistralCallKwargs(MistralCallParams, BaseCallKwargs[dict[str, Any]]):
    model: str
    messages: list[ChatMessage]



from mirascope.core import mistral, prompt_template


@mistral.call("mistral-large-latest", response_model=list[str])
@prompt_template("Extract book titles from {texts}")
def extract_book(texts: list[str]): ...


book = extract_book(
    [
        "The Name of the Wind by Patrick Rothfuss",
        "Mistborn: The Final Empire by Brandon Sanderson",
    ]
)
print(book)
# Output: ["The Name of the Wind", "Mistborn: The Final Empire"]



from mirascope.core import ResponseModelConfigDict, openai
from pydantic import BaseModel


class Book(BaseModel):
    title: str
    author: str

    model_config = ResponseModelConfigDict(strict=True)


@openai.call("gpt-4o-2024-08-06", response_model=Book, json_mode=True)
def recommend_book(genre: str) -> str:
    return f"Recommend a {genre} book"


book = recommend_book("fantasy")
print(book)



from mirascope.core import Messages, prompt_template


@prompt_template()
def recommend_book_prompt(genre: str) -> Messages.Type:
    return Messages.User(f"Recommend a {genre} book")


print(recommend_book_prompt("fantasy"))
# Output: [BaseMessageParam(role='user', content='Recommend a fantasy book')]



from collections.abc import AsyncGenerator, Callable, Generator
from typing import Any, ClassVar, Generic, TypeVar

from pydantic import BaseModel, model_validator
from tenacity import AsyncRetrying, Retrying
from typing_extensions import Self

from ...core.base import BasePrompt, BaseTool, _utils
from ...core.base.dynamic_config import DynamicConfigFull
from .types import BaseCallParams
from .utils import retry_decorator

ExtractedType = _utils.BaseType | BaseModel
_ExtractedTypeT = TypeVar("_ExtractedTypeT", bound=ExtractedType)


class BaseExtractor(BasePrompt, Generic[_ExtractedTypeT]):
    extract_schema: type[_utils.BaseType] | type[BaseModel] | Callable
    tags: ClassVar[list[str]] = []
    prompt_template: ClassVar[str] = ""

    call_params: ClassVar[BaseCallParams] = BaseCallParams(model="")

    _decorator: ClassVar[Callable] = lambda self: None
    _provider: ClassVar[str] = "NO PROVIDER"

    @model_validator(mode="after")
    def set_prompt_template(self) -> Self:
        self._prompt_template = self.prompt_template
        return self

    def dynamic_config(self) -> DynamicConfigFull:
        """Returns the dynamic configuration set from user provided information."""
        config: DynamicConfigFull = {
            "call_params": self.call_params.model_dump(exclude={"model", "tools"}),
            "tools": [
                tool
                if isinstance(tool, type)
                else _utils.convert_function_to_base_tool(tool, BaseTool)
                for tool in self.call_params.tools or []
            ],
        }
        if hasattr(self, "messages"):
            config["messages"] = self.messages()  # pyright: ignore [reportAttributeAccessIssue]
        return config

    def dump(
        self,
    ) -> dict[str, Any]:
        """Dumps the contents of the prompt into a dictionary."""
        return {
            "tags": self.tags,
            "template": _utils.get_prompt_template(self),
            "inputs": self.model_dump(),
        }

    def extract(
        self,
        retries: int | Retrying = 0,
        client: Any | None = None,  # noqa: ANN401
        **kwargs: Any,  # noqa: ANN401
    ) -> _ExtractedTypeT:
        """Extracts the `extract_schema` format from the response."""
        return self.run(
            self.__class__._decorator(
                self.call_params.model,
                response_model=self.extract_schema,
                client=client,
            ),
            retry_decorator(retries),
        )

    async def extract_async(
        self,
        retries: int | Retrying = 0,
        client: Any | None = None,  # noqa: ANN401
        **kwargs: Any,  # noqa: ANN401
    ) -> _ExtractedTypeT:
        """Asynchronously extracts the `extract_schema` format from the response."""
        return await self.run_async(
            self.__class__._decorator(
                self.call_params.model,
                response_model=self.extract_schema,
                client=client,
            ),
            retry_decorator(retries),
        )

    def stream(
        self,
        retries: int | Retrying = 0,
        client: Any | None = None,  # noqa: ANN401
        **kwargs: Any,  # noqa: ANN401
    ) -> Generator[_ExtractedTypeT, None, None]:
        """Streams the desired structured output from the response."""
        yield from self.run(
            self.__class__._decorator(
                self.call_params.model,
                response_model=self.extract_schema,
                stream=True,
                client=client,
            ),
            retry_decorator(retries),
        )

    async def stream_async(
        self,
        retries: int | AsyncRetrying = 0,
        client: Any | None = None,  # noqa: ANN401
        **kwargs: Any,  # noqa: ANN401
    ) -> AsyncGenerator[_ExtractedTypeT, None]:
        """Asynchronously streams the desired structured output from the response."""
        async for item in await self.run_async(
            self.__class__._decorator(
                self.call_params.model,
                response_model=self.extract_schema,
                stream=True,
                client=client,
            ),
            retry_decorator(retries),
        ):
            yield item



from mirascope.core import gemini


@gemini.call("gemini-1.5-flash")
def recommend_book(genre: str) -> str:
    return f"Recommend a {genre} book"


response = recommend_book("fantasy")
print(response.content)



from typing import Annotated

import pytest
from pydantic import BaseModel

from mirascope.core.base._utils._get_fields_from_call_args import (
    get_fields_from_call_args,
)
from mirascope.core.base.from_call_args import FromCallArgs


def test_get_fields_from_get_args_non_pydantic_model():
    response_model = object()

    def fn(x): ...

    result = get_fields_from_call_args(response_model, fn, (), {})
    assert result == {}


def test_get_fields_from_get_args_no_from_call_args():
    class ResponseModel(BaseModel):
        field1: int
        field2: str

    def dummy_fn(field1, field2): ...

    result = get_fields_from_call_args(ResponseModel, dummy_fn, (), {})
    assert result == {}


def test_get_fields_from_get_args_with_from_call_args():
    class ResponseModel(BaseModel):
        field1: Annotated[int, FromCallArgs()]
        field2: str

    def dummy_fn(field1, field2): ...

    result = get_fields_from_call_args(
        ResponseModel, dummy_fn, (10,), {"field2": "test"}
    )
    assert result == {"field1": 10}


def test_get_fields_from_get_args_missing_function_args():
    class ResponseModel(BaseModel):
        field1: Annotated[int, FromCallArgs()]
        field2: str

    def dummy_fn(field2): ...

    with pytest.raises(ValueError) as exc_info:
        get_fields_from_call_args(ResponseModel, dummy_fn, (), {"field2": "test"})
    assert (
        "The function arguments do not contain all the fields marked with `FromCallArgs`"
        in str(exc_info.value)
    )


def test_get_fields_from_get_args_non_callable_fn():
    class ResponseModel(BaseModel):
        field1: Annotated[int, FromCallArgs()]

    fn = None  # Non-callable object
    with pytest.raises(TypeError):
        get_fields_from_call_args(ResponseModel, fn, (), {})  # type: ignore


def test_get_fields_from_get_args_validation_error():
    class ResponseModel(BaseModel):
        field1: Annotated[int, FromCallArgs()]

    def dummy_fn(field2): ...

    with pytest.raises(ValueError) as exc_info:
        get_fields_from_call_args(ResponseModel, dummy_fn, (), {"field2": "test"})
    assert (
        "The function arguments do not contain all the fields marked with `FromCallArgs`"
        in str(exc_info.value)
    )


def test_get_fields_from_get_args_multiple_from_call_args():
    class ResponseModel(BaseModel):
        field1: Annotated[int, FromCallArgs()]
        field2: Annotated[str, FromCallArgs()]
        field3: float

    def dummy_fn(field1, field2, field3): ...

    result = get_fields_from_call_args(
        ResponseModel, dummy_fn, (10,), {"field2": "test", "field3": 3.14}
    )
    assert result == {"field1": 10, "field2": "test"}


def test_get_fields_from_get_args_args_and_kwargs():
    class ResponseModel(BaseModel):
        field1: Annotated[int, FromCallArgs()]
        field2: Annotated[str, FromCallArgs()]

    def dummy_fn(field1, field2): ...

    result = get_fields_from_call_args(
        ResponseModel, dummy_fn, (10,), {"field2": "test"}
    )
    assert result == {"field1": 10, "field2": "test"}



"""Get JSON output from a Vertex response."""

import json

from proto.marshal.collections import RepeatedComposite

from ..call_response import VertexCallResponse
from ..call_response_chunk import VertexCallResponseChunk


def get_json_output(
    response: VertexCallResponse | VertexCallResponseChunk, json_mode: bool
) -> str:
    """Extracts the JSON output from a Vertex response."""
    if isinstance(response, VertexCallResponse):
        if json_mode and response.content:
            return response.content
        elif tool_calls := [
            part.function_call
            for candidate in response.response.candidates
            for part in candidate.content.parts
            if part.function_call.args
        ]:
            return json.dumps(
                {
                    k: v if not isinstance(v, RepeatedComposite) else list(v)
                    for k, v in tool_calls[0].args.items()
                }
            )
        else:
            raise ValueError("No tool call or JSON object found in response.")
    elif not json_mode:
        raise ValueError("Vertex only supports structured streaming in json mode.")
    return response.content



from azure.ai.inference import ChatCompletionsClient
from azure.ai.inference.models import ChatRequestMessage
from azure.core.credentials import AzureKeyCredential

client = ChatCompletionsClient(
    endpoint="YOUR_ENDPOINT", credential=AzureKeyCredential("YOUR_KEY")
)


def recommend_book(genre: str) -> str:
    completion = client.complete(
        model="gpt-4o-mini",
        messages=[
            ChatRequestMessage({"role": "user", "content": f"Recommend a {genre} book"})
        ],
    )
    message = completion.choices[0].message
    return message.content if message.content is not None else ""


output = recommend_book("fantasy")
print(output)



from groq import Groq
from mirascope.core import groq, prompt_template


@groq.call("llama-3.1-70b-versatile", client=Groq())
@prompt_template("Recommend a {genre} book")
def recommend_book(genre: str): ...



from mirascope.core import anthropic


@anthropic.call("claude-3-5-sonnet-20240620")
def recommend_book(genre: str) -> anthropic.AnthropicDynamicConfig:
    return {"messages": [{"role": "user", "content": f"Recommend a {genre} book"}]}


response = recommend_book("fantasy")
print(response.content)



from mirascope.core import Messages, prompt_template


@prompt_template()
def recommend_book(genre: str, price: float) -> Messages.Type:
    return Messages.User(f"Recommend a {genre} book under ${price:.2f}")


print(recommend_book("fantasy", 12.3456))
# Output: [BaseMessageParam(role='user', content='Recommend a fantasy book under $12.35')]



from cohere.types.chat_message import ChatMessage
from mirascope.core import cohere


@cohere.call("command-r-plus")
def recommend_book(genre: str) -> cohere.CohereDynamicConfig:
    return {"messages": [ChatMessage(role="user", message=f"Recommend a {genre} book")]}  # pyright: ignore [reportCallIssue, reportReturnType]


response = recommend_book("fantasy")
print(response.content)



from mirascope.core import BaseMessageParam, litellm
from tenacity import retry, stop_after_attempt, wait_exponential


def get_book_author(title: str) -> str:
    if title == "The Name of the Wind":
        return "Patrick Rothfuss"
    elif title == "Mistborn: The Final Empire":
        return "Brandon Sanderson"
    else:
        return "Unknown"


@litellm.call("gpt-4o-mini", tools=[get_book_author])
def identify_author(book: str) -> list[BaseMessageParam]:
    return [BaseMessageParam(role="user", content=f"Who wrote {book}?")]


@retry(
    stop=stop_after_attempt(3),
    wait=wait_exponential(multiplier=1, min=4, max=10),
)
def run():
    response = identify_author("The Name of the Wind")
    if tool := response.tool:
        print(tool.call())
        print(f"Original tool call: {tool.tool_call}")
    else:
        print(response.content)


run()



from mirascope.core import BaseTool, azure, prompt_template
from pydantic import Field


class GetBookAuthor(BaseTool):
    """Returns the author of the book with the given title."""

    title: str = Field(..., description="The title of the book.")

    def call(self) -> str:
        if self.title == "The Name of the Wind":
            return "Patrick Rothfuss"
        elif self.title == "Mistborn: The Final Empire":
            return "Brandon Sanderson"
        else:
            return "Unknown"


@azure.call("gpt-4o-mini", tools=[GetBookAuthor], stream=True)
@prompt_template("Who wrote {book}?")
def identify_authors(books: list[str]): ...


stream = identify_authors(["The Name of the Wind", "Mistborn: The Final Empire"])
for chunk, tool in stream:
    if tool:
        print(tool.call())
    else:
        print(chunk.content, end="", flush=True)



from mirascope.core import Messages, groq


@groq.call("llama-3.1-70b-versatile", stream=True)
def recommend_book(genre: str) -> Messages.Type:
    return Messages.User(f"Recommend a {genre} book")


stream = recommend_book("fantasy")
for chunk, _ in stream:
    print(chunk.content, end="", flush=True)

print(f"Content: {stream.content}")

call_response = stream.construct_call_response()
print(f"Usage: {call_response.usage}")



from azure.ai.inference.models import UserMessage
from mirascope.core import azure


@azure.call("gpt-4o-mini")
def recommend_book(genre: str) -> azure.AzureDynamicConfig:
    return {"messages": [UserMessage(content=f"Recommend a {genre} book")]}


response = recommend_book("fantasy")
print(response.content)



from mirascope.core import prompt_template, vertex
from pydantic import BaseModel, Field


class Eval(BaseModel):
    reasoning: str = Field(..., description="Reasoning for score in  < 100 characters.")
    score: float = Field(..., description="A score between [0, 5]")


@vertex.call("gemini-1.5-flash", response_model=Eval)
@prompt_template(
    """
    Text is toxic if it contains content that is:
    - Harmful, offensive, disrespectful, or promotes negativity
    - Aggressive, demeaning, bigoted, or excessively critical
    - Encouraging or glorifying violence, illegal activities, or self-harm
    - Reinforcing harmful stereotypes or biases

    Use the following scoring criteria:
    0 - No toxicity
    1 - Mild toxicity; minor offensive elements or insensitivity
    2 - Moderate toxicity; some harmful or disrespectful language
    3 - High toxicity; explicit and aggressive harmful content
    4 - Severe toxicity; dangerous, hateful, or extremist content
    5 - Extreme toxicity; could cause significant harm

    Provide concise and clear reasoning for your assigned score.

    Text to evaluate: {text}
    """
)
def evaluate_toxicity(text: str): ...


# Toxic Example
response = evaluate_toxicity(
    "Why even bother trying? With your laziness and abilities, it's probably not even possible anyway."
)
print(response)
# Output: reasoning="Uses demeaning language and dismisses someone's efforts, showing disrespect." score=2.0

# Not Toxic Example
response = evaluate_toxicity(
    "You can do it! Even if it seems hard now, there's always a way."
)
print(response)
# Output: reasoning='The text is positive and supportive, with no harmful elements.' score=0.0



from unittest.mock import MagicMock, patch

from mirascope.integrations.otel import _utils
from mirascope.integrations.otel._with_otel import with_otel


@patch(
    "mirascope.integrations.otel._with_otel.middleware_factory",
    new_callable=MagicMock,
)
def test_with_otel(mock_middleware_factory: MagicMock) -> None:
    """Tests the `with_otel` decorator."""
    mock_fn = MagicMock()
    with_otel()(mock_fn)
    mock_middleware_factory.assert_called_once()
    call_args = mock_middleware_factory.call_args[1]
    assert call_args["custom_context_manager"] == _utils.custom_context_manager
    assert call_args["handle_response_model"] == _utils.handle_response_model
    assert (
        call_args["handle_response_model_async"] == _utils.handle_response_model_async
    )
    assert call_args["handle_call_response"] == _utils.handle_call_response
    assert call_args["handle_call_response_async"] == _utils.handle_call_response_async
    assert call_args["handle_stream"] == _utils.handle_stream
    assert call_args["handle_stream_async"] == _utils.handle_stream_async



from anthropic import AsyncAnthropic
from mirascope.core import anthropic


@anthropic.call("claude-3-5-sonnet-20240620", client=AsyncAnthropic())
async def recommend_book_async(genre: str) -> str:
    return f"Recommend a {genre} book"



from mirascope.core import BaseMessageParam, BaseTool, bedrock


class GetBookAuthor(BaseTool):
    title: str

    def call(self) -> str:
        if self.title == "The Name of the Wind":
            return "Patrick Rothfuss"
        elif self.title == "Mistborn: The Final Empire":
            return "Brandon Sanderson"
        else:
            return "Unknown"


@bedrock.call("anthropic.claude-3-haiku-20240307-v1:0", tools=[GetBookAuthor])
def identify_author(
    book: str, history: list[bedrock.BedrockMessageParam]
) -> list[bedrock.BedrockMessageParam]:
    messages = [*history]
    if book:
        messages.append(BaseMessageParam(role="user", content=f"Who wrote {book}?"))
    return messages


history = []
response = identify_author("The Name of the Wind", history)
history += [response.user_message_param, response.message_param]
while tool := response.tool:
    tools_and_outputs = [(tool, tool.call())]
    history += response.tool_message_params(tools_and_outputs)
    response = identify_author("", history)
    history.append(response.message_param)
print(response.content)
# Output: The Name of the Wind was written by Patrick Rothfuss.



"""DEPRECATED."""

import inspect
import warnings

from ..retries.tenacity import collect_errors

warnings.warn(
    inspect.cleandoc("""
    The import `mirascope.integrations.tenacity` is deprecated in versions `>=1.3`.
    Please instead use the `mirascope.retries.tenacity` import.
    """)
)
__all__ = ["collect_errors"]



import logfire
from mirascope.core import BaseMessageParam, azure
from mirascope.integrations.logfire import with_logfire
from pydantic import BaseModel

logfire.configure()


class Book(BaseModel):
    title: str
    author: str


@with_logfire()
@azure.call("gpt-4o-mini", response_model=Book)
def recommend_book(genre: str) -> list[BaseMessageParam]:
    return [BaseMessageParam(role="user", content=f"Recommend a {genre} book")]


print(recommend_book("fantasy"))



try:
    from cohere import StreamEndStreamedChatResponse, StreamStartStreamedChatResponse
    from cohere.types import (
        TextGenerationStreamedChatResponse,
        ToolCallsGenerationStreamedChatResponse,
    )
except ImportError:  # pragma: no cover
    # When cohere version is less than 5.11.0
    from cohere import (
        StreamedChatResponse_StreamEnd as StreamEndStreamedChatResponse,  # pyright: ignore [reportAttributeAccessIssue]  # noqa: F401
    )
    from cohere import (
        StreamedChatResponse_StreamStart as StreamStartStreamedChatResponse,  # pyright: ignore [reportAttributeAccessIssue]  # noqa: F401
    )
    from cohere.types import (
        StreamedChatResponse_TextGeneration as TextGenerationStreamedChatResponse,  # pyright: ignore [reportAttributeAccessIssue]  # noqa: F401
    )
    from cohere.types import (
        StreamedChatResponse_ToolCallsGeneration as ToolCallsGenerationStreamedChatResponse,  # pyright: ignore [reportAttributeAccessIssue]  # noqa: F401
    )



"""This module contains the `VertexCallResponse` class.

usage docs: learn/calls.md#handling-responses
"""

from google.cloud.aiplatform_v1beta1.types import (
    FunctionResponse,
    GenerateContentResponse,
)
from pydantic import computed_field
from vertexai.generative_models import Content, GenerationResponse, Tool

from ..base import BaseCallResponse
from ._utils import calculate_cost
from .call_params import VertexCallParams
from .dynamic_config import VertexDynamicConfig
from .tool import VertexTool


class VertexCallResponse(
    BaseCallResponse[
        GenerationResponse,
        VertexTool,
        Tool,
        VertexDynamicConfig,
        Content,
        VertexCallParams,
        Content,
    ]
):
    """A convenience wrapper around the Vertex AI `GenerateContentResponse`.

    When calling the Vertex AI API using a function decorated with `vertex_call`, the
    response will be a `VertexCallResponse` instance with properties that allow for
    more convenient access to commonly used attributes.

    Example:

    ```python
    from mirascope.core import prompt_template
    from mirascope.core.vertex import vertex_call


    @vertex_call("gemini-1.5-flash")
    def recommend_book(genre: str) -> str:
        return f"Recommend a {genre} book"


    response = recommend_book("fantasy")  # response is an `VertexCallResponse` instance
    print(response.content)
    ```
    """

    _provider = "vertex"

    @property
    def content(self) -> str:
        """Returns the contained string content for the 0th choice."""
        return self.response.candidates[0].content.parts[0].text

    @property
    def finish_reasons(self) -> list[str]:
        """Returns the finish reasons of the response."""
        finish_reasons = [
            "FINISH_REASON_UNSPECIFIED",
            "STOP",
            "MAX_TOKENS",
            "SAFETY",
            "RECITATION",
            "OTHER",
        ]

        return [
            finish_reasons[candidate.finish_reason]
            for candidate in self.response.candidates
        ]

    @property
    def model(self) -> str:
        """Returns the model name.

        vertex does not return model, so we return the model provided by
        the user.
        """
        return self._model

    @property
    def id(self) -> str | None:
        """Returns the id of the response.

        vertex does not return an id
        """
        return None

    @property
    def usage(self) -> GenerateContentResponse.UsageMetadata:
        """Returns the usage of the chat completion."""
        return self.response.usage_metadata

    @property
    def input_tokens(self) -> int:
        """Returns the number of input tokens."""
        return self.usage.prompt_token_count

    @property
    def output_tokens(self) -> int:
        """Returns the number of output tokens."""
        return self.usage.candidates_token_count

    @property
    def cost(self) -> float | None:
        """Returns the cost of the call."""
        return calculate_cost(self.input_tokens, self.output_tokens, self.model)

    @computed_field
    @property
    def message_param(self) -> Content:
        """Returns the models's response as a message parameter."""
        return Content(role="model", parts=self.response.candidates[0].content.parts)

    @computed_field
    @property
    def tools(self) -> list[VertexTool] | None:
        """Returns the list of tools for the 0th candidate's 0th content part."""
        if self.tool_types is None:
            return None

        extracted_tools = []
        for part in self.response.candidates[0].content.parts:
            tool_call = part.function_call
            for tool_type in self.tool_types:
                if tool_call.name == tool_type._name():
                    extracted_tools.append(tool_type.from_tool_call(tool_call))
                    break

        return extracted_tools

    @computed_field
    @property
    def tool(self) -> VertexTool | None:
        """Returns the 0th tool for the 0th candidate's 0th content part.

        Raises:
            ValidationError: if the tool call doesn't match the tool's schema.
        """
        tools = self.tools
        if tools:
            return tools[0]
        return None

    @classmethod
    def tool_message_params(
        cls, tools_and_outputs: list[tuple[VertexTool, object]]
    ) -> list[FunctionResponse]:
        """Returns the tool message parameters for tool call results.

        Args:
            tools_and_outputs: The list of tools and their outputs from which the tool
                message parameters should be constructed.

        Returns:
            The list of constructed `FunctionResponse` parameters.
        """
        return [
            FunctionResponse(name=tool._name(), response={"result": output})
            for tool, output in tools_and_outputs
        ]



import inspect

from mirascope.core import Messages, prompt_template


@prompt_template()
def recommend_book_prompt(genre: str) -> Messages.Type:
    return Messages.User(
        inspect.cleandoc(
            f"""
            Recommend a {genre} book.
            Output in the format Title by Author.
            """
        )
    )


print(recommend_book_prompt("fantasy"))
# Output: [BaseMessageParam(role='system', content='Recommend a fantasy book.\nOutput in the format Title by Author.')]



"""This module contains the `fn_is_async` function."""

import inspect
from collections.abc import Awaitable, Callable
from typing import TypeVar

from typing_extensions import TypeIs

_R = TypeVar("_R")


def fn_is_async(
    fn: Callable[..., Awaitable[_R]] | Callable[..., _R],
) -> TypeIs[Callable[..., Awaitable[_R]]]:
    return inspect.iscoroutinefunction(fn)



"""Tests the `bedrock.call` module."""

import sys
from unittest.mock import MagicMock, patch

from mirascope.core.bedrock import _utils
from mirascope.core.bedrock.call_params import BedrockCallParams
from mirascope.core.bedrock.call_response import BedrockCallResponse
from mirascope.core.bedrock.call_response_chunk import BedrockCallResponseChunk
from mirascope.core.bedrock.stream import BedrockStream
from mirascope.core.bedrock.tool import BedrockTool


def test_bedrock_call() -> None:
    """Tests the `bedrock_call` decorator."""

    if "mirascope.core.bedrock._call" in sys.modules:
        del sys.modules["mirascope.core.bedrock._call"]

    with patch(
        "mirascope.core.base.call_factory", new_callable=MagicMock
    ) as mock_call_factory:
        import mirascope.core.bedrock._call  # noqa: F401

        mock_call_factory.assert_called_once_with(
            TCallResponse=BedrockCallResponse,
            TCallResponseChunk=BedrockCallResponseChunk,
            TStream=BedrockStream,
            TToolType=BedrockTool,
            default_call_params=BedrockCallParams(),
            setup_call=_utils.setup_call,
            get_json_output=_utils.get_json_output,
            handle_stream=_utils.handle_stream,
            handle_stream_async=_utils.handle_stream_async,
        )



from mirascope.core import prompt_template


@prompt_template(
    "I just read this book: {previous_book:image} What should I read next?"
)
def recommend_book_prompt(previous_book: bytes): ...


print(recommend_book_prompt(b"..."))
# Output: [
#     BaseMessageParam(
#         role="user",
#         content=[
#             TextPart(type="text", text="I just read this book:"),
#             ImagePart(type="image", media_type="image/jpeg", image=b"...", detail=None),
#             TextPart(type="text", text="What should I read next?"),
#         ],
#     )
# ]



from mirascope.core import prompt_template


@prompt_template(
    """
    SYSTEM: You are a librarian
    USER: Recommend a {genre} book
    """
)
def recommend_book_prompt(genre: str): ...


print(recommend_book_prompt("fantasy"))
# Output: [
#   BaseMessageParam(role='system', content='You are a librarian'),
#   BaseMessageParam(role='user', content='Recommend a fantasy book'),
# ]



import json

from mirascope.core import anthropic


@anthropic.call("claude-3-5-sonnet-20240620", json_mode=True)
def get_book_info(book_title: str) -> str:
    return f"Provide the author and genre of {book_title}"


response = get_book_info("The Name of the Wind")
print(json.loads(response.content))
# Output: {'author': 'Patrick Rothfuss', 'genre': 'Fantasy'}



from unittest.mock import MagicMock, patch

from mirascope.integrations.langfuse._with_langfuse import (
    custom_decorator,
    with_langfuse,
)


@patch(
    "mirascope.integrations.langfuse._with_langfuse.middleware_factory",
    new_callable=MagicMock,
)
@patch(
    "mirascope.integrations.langfuse._with_langfuse.observe",
    new_callable=MagicMock,
)
def test_with_langfuse(
    mock_observe: MagicMock, mock_middleware_factory: MagicMock
) -> None:
    """Tests the `with_langfuse` decorator."""
    mock_fn = MagicMock(__name__="mock_fn")
    mock_observe.return_value = MagicMock()
    with_langfuse()(mock_fn)
    mock_middleware_factory.assert_called_once()
    call_args = mock_middleware_factory.call_args[1]
    assert call_args["custom_decorator"] is not None
    assert call_args["handle_response_model"] is not None
    assert call_args["handle_response_model_async"] is not None
    assert call_args["handle_call_response"] is not None
    assert call_args["handle_call_response_async"] is not None
    assert call_args["handle_stream"] is not None
    assert call_args["handle_stream_async"] is not None


@patch(
    "mirascope.integrations.langfuse._with_langfuse.observe",
    new_callable=MagicMock,
)
def test_custom_decorator(mock_observe: MagicMock) -> None:
    """Tests the `custom_decorator` function."""
    mock_fn = MagicMock(__name__="mock_fn")
    custom_decorator(mock_fn)
    mock_observe.assert_called_once_with(
        name=mock_fn.__name__,
        as_type="generation",
        capture_input=False,
        capture_output=False,
    )



"""The `OpenAITool` class for easy tool usage with OpenAI LLM calls.

usage docs: learn/tools.md
"""

from __future__ import annotations

import jiter
from openai.types.chat import (
    ChatCompletionMessageToolCall,
    ChatCompletionToolParam,
)
from openai.types.shared_params import FunctionDefinition
from pydantic.json_schema import SkipJsonSchema

from ..base import BaseTool, GenerateJsonSchemaNoTitles, ToolConfig


class GenerateOpenAIStrictToolJsonSchema(GenerateJsonSchemaNoTitles):
    _openai_strict = True


class OpenAIToolConfig(ToolConfig, total=False):
    """A tool configuration for OpenAI-specific features."""

    strict: bool


class OpenAITool(BaseTool[ChatCompletionToolParam]):
    """A class for defining tools for OpenAI LLM calls.

    Example:

    ```python
    from mirascope.core import prompt_template
    from mirascope.core.openai import openai_call


    def format_book(title: str, author: str) -> str:
        return f"{title} by {author}"


    @openai_call("gpt-4o-mini", tools=[format_book])
    def recommend_book(genre: str) -> str:
        return f"Recommend a {genre} book"


    response = recommend_book("fantasy")
    if tool := response.tool:  # returns an `OpenAITool` instance
        print(tool.call())
    ```
    """

    __provider__ = "openai"
    __tool_config_type__ = OpenAIToolConfig

    tool_call: SkipJsonSchema[ChatCompletionMessageToolCall]

    @classmethod
    def tool_schema(cls) -> ChatCompletionToolParam:
        """Constructs a JSON Schema tool schema from the `BaseModel` schema defined.

        Example:
        ```python
        from mirascope.core.openai import OpenAITool


        def format_book(title: str, author: str) -> str:
            return f"{title} by {author}"


        tool_type = OpenAITool.type_from_fn(format_book)
        print(tool_type.tool_schema())  # prints the OpenAI-specific tool schema
        ```
        """
        fn = FunctionDefinition(name=cls._name(), description=cls._description())
        schema_generator = GenerateJsonSchemaNoTitles
        if cls.tool_config.get("strict", False):
            fn["strict"] = True
            schema_generator = GenerateOpenAIStrictToolJsonSchema
        model_schema = cls.model_json_schema(schema_generator=schema_generator)
        if model_schema["properties"]:
            fn["parameters"] = model_schema
        return ChatCompletionToolParam(function=fn, type="function")

    @classmethod
    def from_tool_call(cls, tool_call: ChatCompletionMessageToolCall) -> OpenAITool:
        """Constructs an `OpenAITool` instance from a `tool_call`.

        Args:
            tool_call: The OpenAI tool call from which to construct this tool instance.
        """
        model_json = jiter.from_json(tool_call.function.arguments.encode())
        model_json["tool_call"] = tool_call.model_dump()
        return cls.model_validate(model_json)



import re

from mirascope.core import openai, prompt_template


def parse_cot(response: openai.OpenAICallResponse) -> str:
    pattern = r"<thinking>.?*</thinking>.*?<output>(.*?)</output>"
    match = re.search(pattern, response.content, re.DOTALL)
    if not match:
        return response.content
    return match.group(1).strip()


@openai.call("gpt-4o-mini", output_parser=parse_cot)
@prompt_template(
    """
    First, output your thought process in <thinking> tags.
    Then, provide your final output in <output> tags.

    Question: {question}
    """
)
def chain_of_thought(question: str): ...


question = "Roger has 5 tennis balls. He buys 2 cans of 3. How many does he have now?"
output = chain_of_thought(question)
print(output)



from typing import Annotated, cast

from mirascope.core import cohere, prompt_template
from pydantic import AfterValidator, BaseModel, ValidationError


def validate_upper(v: str) -> str:
    assert v.isupper(), "Field must be uppercase"
    return v


class Book(BaseModel):
    """An extracted book."""

    title: Annotated[str, AfterValidator(validate_upper)]
    author: Annotated[str, AfterValidator(validate_upper)]


@cohere.call("command-r-plus", response_model=Book)
@prompt_template("Extract {text}")
def extract_book(text: str): ...


try:
    book = extract_book("The Name of the Wind by Patrick Rothfuss")
    print(book)
    # Output: title='The Name of the Wind' author='Patrick Rothfuss'
except ValidationError as e:
    print(f"Error: {str(e)}")
    # Error: 2 validation errors for Book
    # title
    #   Assertion failed, Field must be uppercase [type=assertion_error, input_value='The Name of the Wind', input_type=str]
    #     For further information visit https://errors.pydantic.dev/2.7/v/assertion_error
    # author
    #   Assertion failed, Field must be uppercase [type=assertion_error, input_value='Patrick Rothfuss', input_type=str]
    #     For further information visit https://errors.pydantic.dev/2.7/v/assertion_error
    response = cast(cohere.CohereCallResponse, e._response)  # pyright: ignore[reportAttributeAccessIssue]
    print(response.model_dump())
    # > {'metadata': {}, 'response': {'id': ...}, ...}



from groq import AsyncGroq
from mirascope.core import groq


@groq.call("llama-3.1-70b-versatile", client=AsyncGroq())
async def recommend_book_async(genre: str) -> str:
    return f"Recommend a {genre} book"
